import streamlit as st
from gui_elements.stateful_widget import StatefulWidgets

from qstn.inference.response_generation import (
    ChoiceResponseGenerationMethod,
    JSONReasoningResponseGenerationMethod,
    JSONSingleResponseGenerationMethod,
    JSONVerbalizedDistribution,
)
from qstn.prompt_builder import generate_likert_options
from qstn.utilities.prompt_templates import (
    LIST_OPTIONS_DEFAULT,
    SCALE_OPTIONS_DEFAULT,
)

st.set_page_config(layout="wide")

st.title("Likert Scale Options Generator")
st.write(
    "This interface allows you to configure and generate Likert scale answer "
    "options by adjusting the parameters below."
)
st.page_link(
    "pages/00_Tutorial.py",
    label="📖 Need help? See tutorial: Answer options",
    query_params={"section": "options"},
)
st.divider()

if "questionnaires" not in st.session_state:
    st.error("You need to first upload a questionnaire and the population you want to survey.")
    st.stop()
    disabled = True
else:
    disabled = False

# if 'answer_texts_input' not in st.session_state:
# st.session_state.answer_texts_input = (
#     "Strongly Disagree\nDisagree\nNeutral\nAgree\nStrongly Agree"
# )


@st.cache_data
def create_stateful_widget() -> StatefulWidgets:
    return StatefulWidgets()


state = create_stateful_widget()

# Use a form to batch all inputs together
with st.container(border=True):
    # --- Main Configuration ---
    st.subheader("Main Configuration")
    col1, col2, col3 = st.columns(3)

    with col1:
        n_options = state.create(
            st.number_input,
            "n_options",
            "Number of Options (n)",
            initial_value=5,
            min_value=2,
            step=1,
            help="The total number of choices in the scale.",
        )

    with col2:
        idx_type = state.create(
            st.selectbox,
            "idx_type",
            "Index Type",
            initial_value="integer",
            options=["integer", "char_low", "char_up"],
            help="The type of index to use for the options.",
        )

    with col3:
        start_idx = state.create(
            st.number_input,
            "start_idx",
            "Starting Index",
            initial_value=1,
            step=1,
            help="The number to start counting from (e.g., 1).",
        )

    # --- Order and Structure Options ---
    st.subheader("Ordering and Structure")
    col1, col2, col3, col4 = st.columns(4)

    with col1:
        only_from_to_scale = state.create(
            st.checkbox,
            "only_from_to_scale",
            "From-To Scale Only",
            initial_value=False,
            help=(
                "If checked, only the first and last answer labels are display "
                "e.g. 1 Strongly Disagree to 5 Strongly agree."
            ),
        )

    with col2:
        random_order = state.create(
            st.checkbox,
            "random_order",
            "Random Order",
            initial_value=False,
            help="Randomize the order of options.",
        )

    with col3:
        reversed_order = state.create(
            st.checkbox,
            "reversed_order",
            "Reversed Order",
            initial_value=False,
            help="Reverse the order of options.",
        )

    with col4:
        even_order = state.create(
            st.checkbox,
            "even_order",
            "Even Order",
            initial_value=False,
            help=(
                "If there is an uneven number of answer texts, the middle "
                "section is automatically removed."
            ),
        )

    # --- Answer Texts Input ---
    st.subheader("Answer Texts")

    answer_texts = state.create(
        st.text_area,
        "answer_texts",
        "Enter Answer Texts (one per line)",
        initial_value="Strongly Disagree\nDisagree\nNeutral\nAgree\nStrongly Agree",
        height=170,
        help="Enter the labels for each answer option.",
    )

    # --- Advanced Configuration ---
    with st.expander("Advanced Configuration"):
        options_separator = state.create(
            st.text_input,
            "options_separator",
            "Options Separator",
            initial_value=", ",
            help="The character(s) used to separate options in the final string.",
        )
        list_prompt_template = state.create(
            st.text_area,
            "list_prompt_template",
            "List Prompt Template",
            initial_value=LIST_OPTIONS_DEFAULT,
            height=100,
            help="Write how the options should be presented to the model.",
        )
        scale_prompt_template = state.create(
            st.text_area,
            "scale_prompt_template",
            "Scale Prompt Template",
            initial_value=SCALE_OPTIONS_DEFAULT,
            height=100,
            help="Write how the options should be presented to the model.",
        )

    # --- Response Generation Method ---
    st.subheader("Response Generation Method")
    st.write(
        "Choose how the model output should be constrained. This controls the format of responses."
    )

    # Global option for all response generation methods
    output_index_only = state.create(
        st.checkbox,
        "output_index_only",
        "Output Indices Only",
        initial_value=False,
        help=(
            "If checked, the model will output only indices (e.g., '1', '2', "
            "'3') instead of full text (e.g., '1: Strongly Disagree'). "
            "Applies to all response generation methods."
        ),
    )

    rgm_type = state.create(
        st.selectbox,
        "rgm_type",
        "Response Generation Method",
        options=[
            "None",
            "JSON Single Answer",
            "JSON All Options (Probabilities)",
            "JSON with Reasoning",
            "Choice",
        ],  # "Logprob" commented out - not fully implemented
        initial_value="None",
        help=(
            "None: Free text output (requires parsing). JSON: Structured JSON "
            "output. Choice: Guided choice selection."
        ),
    )

    # Show additional options based on selected method
    rgm_config = {}

    # Choice configuration - allow user to specify choices
    if rgm_type == "Choice":
        with st.container(border=True):
            st.write("**Choice Configuration:**")
            # Pre-fill with answer texts if available
            if answer_texts:
                answer_texts_list_preview = [
                    text.strip() for text in answer_texts.split("\n") if text.strip()
                ]
                # Format as full text by default: "1: Option 1, 2: Option 2"
                if output_index_only:
                    # If indices only, just use numbers
                    default_choices = "\n".join(
                        [str(i + 1) for i in range(len(answer_texts_list_preview))]
                    )
                else:
                    # Full text format
                    default_choices = "\n".join(
                        [f"{i + 1}: {text}" for i, text in enumerate(answer_texts_list_preview)]
                    )
            else:
                default_choices = ""

            choice_input = state.create(
                st.text_area,
                "choice_allowed_choices",
                "Allowed Choices (one per line)",
                initial_value=default_choices,
                height=100,
                help=(
                    "Enter the choices the model can output. Format: indices "
                    "only (1, 2, 3) or full text "
                    "(1: Strongly Disagree, 2: Disagree, etc.)."
                ),
            )
            rgm_config["allowed_choices"] = choice_input

    # Logprob configuration - commented out until fully implemented
    # if rgm_type == "Logprob":
    #     with st.container(border=True):
    #         st.write("**Logprob Configuration:**")
    #         rgm_config["token_position"] = state.create(
    #             st.number_input,
    #             "logprob_token_position",
    #             "Token Position",
    #             initial_value=0,
    #             min_value=0,
    #             help="Position in output to capture logprobs (0 = first token).",
    #         )
    #         rgm_config["token_limit"] = state.create(
    #             st.number_input,
    #             "logprob_token_limit",
    #             "Token Limit",
    #             initial_value=1,
    #             min_value=1,
    #             help="Maximum number of tokens to generate.",
    #         )
    #         rgm_config["top_logprobs"] = state.create(
    #             st.number_input,
    #             "logprob_top_logprobs",
    #             "Top Logprobs",
    #             initial_value=20,
    #             min_value=1,
    #             max_value=20,
    #             help="Number of top probabilities to return (max 20).",
    #         )
    #         rgm_config["ignore_reasoning"] = state.create(
    #             st.checkbox,
    #             "logprob_ignore_reasoning",
    #             "Ignore Reasoning",
    #             initial_value=True,
    #             help="Only consider tokens after reasoning output.",
    #         )

    # The submit button for the form
    submitted = st.button(
        "Confirm and Generate Options",
        disabled=disabled,
        type="primary",
        use_container_width=True,
    )

    if st.button("Skip", use_container_width=True, icon="❌"):
        st.session_state.survey_options = None
        st.switch_page("pages/02_Prompt_Configuration.py")

# --- Processing and Output ---
if submitted:
    # print("Session state answer texts "+ st.session_state.answer_texts_input)
    # print(answer_texts_input)
    # Convert the raw text area string into a list of strings.
    answer_texts_list = [text.strip() for text in answer_texts.split("\n") if text.strip()]

    # --- Input Validation ---
    validation_ok = True
    if only_from_to_scale and len(answer_texts_list) != 2:
        st.error(
            "Error: When 'From-To Scale Only' is selected, you must provide "
            f"exactly 2 answer texts. You provided {len(answer_texts_list)}."
        )
        validation_ok = False

    if not only_from_to_scale and len(answer_texts_list) != n_options:
        st.error(
            "Error: The number of answer texts "
            f"({len(answer_texts_list)}) must match the 'Number of Options' "
            f"({n_options})."
        )
        validation_ok = False

    if reversed_order and random_order:
        st.error("Error: Reversed Order and Random Order cannot both be true.")
        validation_ok = False

    if validation_ok:
        # Create response generation method based on user selection
        response_generation_method = None
        if rgm_type == "JSON Single Answer":
            response_generation_method = JSONSingleResponseGenerationMethod(
                output_index_only=output_index_only,
            )
        elif rgm_type == "JSON All Options (Probabilities)":
            response_generation_method = JSONVerbalizedDistribution(
                output_index_only=output_index_only,
            )
        elif rgm_type == "JSON with Reasoning":
            response_generation_method = JSONReasoningResponseGenerationMethod(
                output_index_only=output_index_only,
            )
        elif rgm_type == "Choice":
            # Parse the choices from the text input
            choice_text = rgm_config.get("allowed_choices", "")
            if choice_text:
                # Split by newlines and clean up
                allowed_choices_list = [
                    choice.strip() for choice in choice_text.split("\n") if choice.strip()
                ]
                choice_kwargs = {"allowed_choices": allowed_choices_list}
            else:
                choice_kwargs = {"allowed_choices_template": "{options}"}

            response_generation_method = ChoiceResponseGenerationMethod(
                output_index_only=output_index_only,
                **choice_kwargs,
            )
        # Logprob - commented out until fully implemented
        # elif rgm_type == "Logprob":
        #     response_generation_method = LogprobResponseGenerationMethod(
        #         token_position=rgm_config.get("token_position", 0),
        #         token_limit=rgm_config.get("token_limit", 1),
        #         top_logprobs=rgm_config.get("top_logprobs", 20),
        #         ignore_reasoning=rgm_config.get("ignore_reasoning", True),
        #         automatic_output_instructions=True,
        #         output_index_only=False,
        #     )
        # If rgm_type == "None", response_generation_method stays None

        survey_options = generate_likert_options(
            n=n_options,
            answer_texts=answer_texts_list,
            only_from_to_scale=only_from_to_scale,
            random_order=random_order,
            reversed_order=reversed_order,
            even_order=even_order,
            start_idx=start_idx,
            list_prompt_template=list_prompt_template,
            scale_prompt_template=scale_prompt_template,
            options_separator=options_separator,
            idx_type=idx_type,
            response_generation_method=response_generation_method,
        )

        st.session_state.survey_options = survey_options

        # Auto-save session
        from gui_elements.session_cache import save_session_state

        save_session_state()

        st.switch_page("pages/02_Prompt_Configuration.py")



import random
from typing import TYPE_CHECKING, Any

from tqdm.auto import tqdm

from .response_generation import (
    LogprobResponseGenerationMethod,
    ResponseGenerationMethod,
)
from .utils import normalize_system_messages

if TYPE_CHECKING:
    from openai import AsyncOpenAI
    from vllm import LLM  # pyright: ignore[reportMissingImports]

HAS_VLLM = False
HAS_OPENAI = False

try:
    from vllm import LLM  # pyright: ignore[reportMissingImports]

    from .local_inference import run_vllm_batch, run_vllm_batch_conversation

    HAS_VLLM = True
except ImportError:
    LLM = None

try:
    from openai import AsyncOpenAI

    from .remote_inference import run_openai_batch, run_openai_batch_conversation

    HAS_OPENAI = True
except ImportError:
    AsyncOpenAI = None


def _print_conversation(
    system_messages: list[str | None],
    prompts: list[str],
    assistant_messages: list[str] | None,
    plain_results: list[str],
    reasoning_output: list[str],
    logprob_result: list[str],
    response_generation_method: list[ResponseGenerationMethod] | None,
    number_of_printed_conversations: int = 2,
):
    methods = response_generation_method or []

    if reasoning_output is None:
        reasonings = [None] * len(system_messages)
    else:
        reasonings = reasoning_output

    if logprob_result is None:
        logprobs = [None] * len(system_messages)
    else:
        logprobs = logprob_result

    if assistant_messages:

        conversation_print = "--- Conversation ---"
        for i, (
            system_message,
            prompt_list,
            answer,
            reasoning,
            logprob_answer,
            assistant_list,
        ) in enumerate(
            zip(
                system_messages,
                prompts,
                plain_results,
                reasonings,
                logprobs,
                assistant_messages,
            )
        ):
            if i >= number_of_printed_conversations:
                break

            round_print = conversation_print
            if system_message is not None:
                round_print = f"{round_print}\n-- System Message --\n{system_message}"
            for j, _ in enumerate(prompt_list):
                round_print = f"{round_print}\n-- User Message --\n{prompt_list[j]}"
                if j < len(assistant_list):
                    prefill = assistant_list[j]
                    if prefill:
                        round_print = f"{round_print}\n-- Assistant Message --\n{assistant_list[j]}"
            round_print = f"{round_print}\n-- Generated Answer --\n{answer}"
            if reasoning:
                round_print += "\n-- Reasoning --\n" + str(reasoning)

            if i < len(methods):
                current_method = methods[i]
                if isinstance(current_method, LogprobResponseGenerationMethod):
                    round_print += "\n-- Logprobs --\n" + str(logprob_answer)
            tqdm.write(round_print)
    else:
        conversation_print = "--- Conversation ---"
        for i, (system_message, prompt, answer, reasoning, logprob_answer) in enumerate(
            zip(system_messages, prompts, plain_results, reasonings, logprobs)
        ):
            if i >= number_of_printed_conversations:
                break
            round_print = conversation_print
            if system_message is not None:
                round_print = f"{round_print}\n-- System Message --\n{system_message}"
            round_print = (
                f"{round_print}\n-- User Message ---\n{prompt}"
                f"\n-- Generated Message --\n{answer}"
            )
            if reasoning:
                round_print += "\n-- Reasoning --\n" + str(reasoning)

            if i < len(methods):
                current_method = methods[i]
                if isinstance(current_method, LogprobResponseGenerationMethod):
                    round_print += "\n-- Logprobs --\n" + str(logprob_answer)
            tqdm.write(round_print)


def batch_generation(
    model: LLM | AsyncOpenAI,  # pyright: ignore[reportInvalidTypeForm]
    system_messages: list[str | None] | None = ("You are a helpful assistant.",),
    prompts: list[str] = ("Hi there! What is your name?",),
    response_generation_method: (
        ResponseGenerationMethod | list[ResponseGenerationMethod] | None
    ) = None,
    seed: int = 42,
    client_model_name: str | None = None,
    api_concurrency: int = 10,
    print_conversation: bool = False,
    number_of_printed_conversations: int = 2,
    print_progress: bool = True,
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    space_char: str = "Ġ",
    **generation_kwargs: Any,
) -> tuple:
    """
    Generate responses for a batch of prompts.

    Handles both vLLM and OpenAI API generation with support for:
    - Structured output (JSON or choice format)
    - Conversation printing
    - Progress tracking
    - Concurrent API requests

    Args:
        model (LLM or AsyncOpenAI): vLLM model or AsyncOpenAI client.
        system_messages (List(str)): System prompts for each conversation.
        prompts (List(str)): User prompts to generate responses for.
        response_generation_method (
            ResponseGenerationMethod or List(ResponseGenerationMethod), optional
        ): Configuration for structured output.
        seed (int): Random seed for reproducibility, defaults to 42.
        client_model_name (str, optional): Model name when using OpenAI API.
        api_concurrency (int): Max concurrent API requests when using OpenAI API.
        number_if_printed_conversations (int): How many conversations should be
            printed. Defaults to 2.
        print_conversation (bool): If True, prints conversations. Defaults to False.
        print_progress (bool): If True, shows progress bar. Defaults to True.
        reasoning_start_token (str): Special token at the beginning of reasoning
            models' output. Used for manual parsing if automatic parsing fails.
        reasoning_end_token (str): Special token to separate reasoning from
            regular model output. Used for manual parsing if automatic parsing
            fails.
        space_token (str): Special char to encode spaces in tokens ("Ġ" for
            most byte-pair tokenizers).
        generation_kwargs: Additional generation parameters

    Returns:
        Tuple[List[str], List[str], List[str]]: Generated Response, Logprobs, Reasoning
    """

    model_type = type(model).__name__
    if model_type == "LLM" and not HAS_VLLM:
        raise ImportError("You are trying to use a vLLM model, but 'vllm' is not installed.")
    if model_type == "AsyncOpenAI" and not HAS_OPENAI:
        raise ImportError("You are trying to use OpenAI, but 'openai' is not installed.")
    if model_type != "LLM" and model_type != "AsyncOpenAI":
        raise ValueError(f"Unsupported model type: {type(model)}")
    random.seed(seed)

    normalized_system_messages = normalize_system_messages(
        system_messages=system_messages,
        batch_size=len(prompts),
    )

    # Inference
    if HAS_VLLM and isinstance(model, LLM):
        plain_results, logprob_result, reasoning_outputs = run_vllm_batch(
            model,
            system_messages=normalized_system_messages,
            prompts=prompts,
            response_generation_method=response_generation_method,
            seed=seed,
            print_progress=print_progress,
            reasoning_start_token=reasoning_start_token,
            reasoning_end_token=reasoning_end_token,
            space_char=space_char,
            **generation_kwargs,
        )
    elif HAS_OPENAI and isinstance(model, AsyncOpenAI):
        plain_results, logprob_result, reasoning_outputs = run_openai_batch(
            model,
            system_messages=normalized_system_messages,
            prompts=prompts,
            response_generation_method=response_generation_method,
            seed=seed,
            print_progress=print_progress,
            client_model_name=client_model_name,
            api_concurrency=api_concurrency,
            reasoning_start_token=reasoning_start_token,
            reasoning_end_token=reasoning_end_token,
            **generation_kwargs,
        )
    else:
        raise ValueError("Inference cannot be run without OpenAI or vllm installed.")

    if print_conversation:
        _print_conversation(
            system_messages=normalized_system_messages,
            prompts=prompts,
            assistant_messages=None,
            plain_results=plain_results,
            reasoning_output=reasoning_outputs,
            logprob_result=logprob_result,
            response_generation_method=response_generation_method,
            number_of_printed_conversations=number_of_printed_conversations,
        )

    return (plain_results, logprob_result, reasoning_outputs)


def batch_turn_by_turn_generation(
    model: LLM | AsyncOpenAI,  # type: ignore
    system_messages: list[str | None] | None = ("You are a helpful assistant.",),
    prompts: list[list[str]] = (
        (
            "Hi there! What is your name?",
            "Interesting",
        ),
    ),
    assistant_messages: list[list[str]] | None = None,
    response_generation_method: (
        ResponseGenerationMethod | list[ResponseGenerationMethod] | None
    ) = None,
    seed: int = 42,
    client_model_name: str | None = None,
    api_concurrency: int = 10,
    print_conversation: bool = False,
    number_of_printed_conversations: int = 2,
    print_progress: bool = True,
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    space_char: str = "Ġ",
    **generation_kwargs,
) -> list[str]:
    """
    Generate responses for multi-turn conversations.

    Handles conversations with multiple back-and-forth exchanges between
    user and assistant. Supports:
    - Structured output formats
    - Pre-filled assistant messages
    - Conversation printing
    - Progress tracking

    Args:
        model (LLM or AsyncOpenAI): vLLM model or AsyncOpenAI client.
        system_messages (List(str)): System prompts for each conversation.
        prompts (List(List(str))): User prompts to generate responses for. Can
            include multiple requests per system prompt.
        assistant_messages (List(List(str)), optional): Prefilled assistant
            responses. For example, if the first list contains one entry, the
            first assistant turn is prefilled and not inferred.
        response_generation_method (
            ResponseGenerationMethod or List(ResponseGenerationMethod), optional
        ): Configuration for structured output.
        seed (int): Random seed for reproducibility, defaults to 42.
        client_model_name (str, optional): Model name when using OpenAI API.
        api_concurrency (int): Max concurrent API requests when using OpenAI API.
        print_conversation (bool): If True, prints conversations. Defaults to False.
        number_of_printed_conversations (int): How many conversations should be
            printed. Defaults to 2.
        print_progress (bool): If True, shows progress bar. Defaults to True.
        reasoning_start_token (str): Special token at the beginning of reasoning
            models' output. Used for manual parsing if automatic parsing fails.
        reasoning_end_token (str): Special token to separate reasoning from
            regular model output. Used for manual parsing if automatic parsing
            fails.
        space_token (str): Special char to encode spaces in tokens ("Ġ" for
            most byte-pair tokenizers).
        generation_kwargs: Additional generation parameters.

    Returns:
        Tuple[List[str], List[str], List[str]]: Generated Response, Logprobs, Reasoning
    """

    model_type = type(model).__name__
    if model_type == "LLM" and not HAS_VLLM:
        raise ImportError("You are trying to use a vLLM model, but 'vllm' is not installed.")
    elif model_type == "AsyncOpenAI" and not HAS_OPENAI:
        raise ImportError("You are trying to use OpenAI, but 'openai' is not installed.")
    elif model_type != "LLM" and model_type != "AsyncOpenAI":
        raise ValueError(f"Unsupported model type: {type(model)}")
    random.seed(seed)

    normalized_system_messages = normalize_system_messages(
        system_messages=system_messages,
        batch_size=len(prompts),
    )

    # Inference
    if HAS_VLLM and isinstance(model, LLM):
        plain_results, logprob_result, reasoning_outputs = run_vllm_batch_conversation(
            model,
            system_messages=normalized_system_messages,
            prompts=prompts,
            assistant_messages=assistant_messages,
            response_generation_method=response_generation_method,
            seed=seed,
            print_progress=print_progress,
            reasoning_start_token=reasoning_start_token,
            reasoning_end_token=reasoning_end_token,
            space_char=space_char,
            **generation_kwargs,
        )
    elif HAS_OPENAI and isinstance(model, AsyncOpenAI):
        plain_results, logprob_result, reasoning_outputs = run_openai_batch_conversation(
            model,
            system_messages=normalized_system_messages,
            prompts=prompts,
            assistant_messages=assistant_messages,
            response_generation_method=response_generation_method,
            client_model_name=client_model_name,
            seed=seed,
            print_progress=print_progress,
            api_concurrency=api_concurrency,
            **generation_kwargs,
        )
    else:
        raise ValueError("Inference cannot be run without OpenAI or vllm installed.")

    if print_conversation:
        _print_conversation(
            system_messages=normalized_system_messages,
            prompts=prompts,
            assistant_messages=assistant_messages,
            plain_results=plain_results,
            reasoning_output=reasoning_outputs,
            logprob_result=logprob_result,
            response_generation_method=response_generation_method,
            number_of_printed_conversations=number_of_printed_conversations,
        )

    return (plain_results, logprob_result, reasoning_outputs)


# def batch_decoding(
#     model: Union[LLM, AsyncOpenAI],
#     prompts: List[str] = ["Hi there! What is your name?"],
#     stop_tokens: List[str] = ["\nA:"],
#     structured_output_options: Optional[
#         Union[ResponseGenerationMethod, List[ResponseGenerationMethod]]
#     ] = None,
#     seed: int = 42,
#     client_model_name: Optional[str] = None,
#     api_concurrency: int = 10,
#     print_conversation: bool = False,
#     print_progress: bool = True,
#     **generation_kwargs: Any,
# ):
#     """
#     Generate responses for a batch of prompts.

#     Handles both vLLM and OpenAI API generation with support for:
#     - Structured output (JSON or choice format)
#     - Conversation printing
#     - Progress tracking
#     - Concurrent API requests

#     Args:
#         model: vLLM model or AsyncOpenAI client
#         system_messages: System prompts for each conversation
#         prompts: User prompts to generate responses for
#         structured_output_options: Configuration for structured output
#         seed: Random seed for reproducibility
#         client_model_name: Model name when using OpenAI API
#         api_concurrency: Max concurrent API requests
#         print_conversation: If True, prints conversations
#         print_progress: If True, shows progress bar
#         **generation_kwargs: Additional generation parameters

#     Returns:
#         List[str]: Generated responses
#     """
#     random.seed(seed)

#     batch_size: int = len(prompts)

#     seeds = _generate_seeds(seed, batch_size=batch_size)

#     if isinstance(model, LLM):
#         sampling_params_list = _create_sampling_params(
#             batch_size=batch_size,
#             seeds=seeds,
#             structured_output_options=structured_output_options,
#             stop_tokens=stop_tokens,
#             **generation_kwargs,
#         )
#         outputs: List[RequestOutput] = model.generate(
#             prompts,
#             sampling_params=sampling_params_list,
#             use_tqdm=print_progress,
#         )
#         result = [output.outputs[0].text for output in outputs]

#     else:
#         result = _run_async_in_thread(
#             client=model,
#             client_model_name=client_model_name,
#             batch_messages=prompts,
#             seeds=seeds,
#             concurrency_limit=api_concurrency,
#             structured_output_options=structured_output_options,
#             **generation_kwargs,
#         )

#     # TODO add argument to specify how many conversations should be printed.
#     # The base argument should be reasonable.
#     if print_conversation:
#         conversation_print = "Conversation:"
#         for prompt, answer in zip(prompts, result):
#             round_print = (
#                 f"{conversation_print}\nUser Message:\n{prompt}"
#                 f"\nGenerated Message\n{answer}"
#             )
#             print(round_print, flush=True)
#             break

#     return result



"""Notebook-style integration tests for the German-persona workflow.

These tests mirror the data flow in `docs/resources/german_general_personas.ipynb`
while mocking heavy inference calls.
"""

import json

import pandas as pd

from qstn import survey_manager
from qstn.inference import response_generation
from qstn.parser import parse_json, raw_responses
from qstn.prompt_builder import LLMPrompt, generate_likert_options
from qstn.utilities import create_one_dataframe, placeholder


def _build_notebook_like_inputs():
    personas = pd.DataFrame(
        [
            {"persona": "Person A: lebt in Berlin."},
            {"persona": "Person B: lebt in Hamburg."},
        ]
    )

    json_questionnaire = {
        "mp18": {
            "statement": "Wie zufrieden sind Sie mit der Demokratie?",
            "answers": ["1: Sehr unzufrieden", "2: Sehr zufrieden"],
        },
        "pizza": {
            "statement": "Wie stehen Sie zu Ananas auf Pizza?",
            "answers": ["1: Schrecklich", "2: Lecker"],
        },
    }

    questionnaire = pd.DataFrame(
        [
            {"questionnaire_item_id": qid, "question_content": payload["statement"]}
            for qid, payload in json_questionnaire.items()
        ]
    )

    all_cleaned_answers = []
    for qid, payload in json_questionnaire.items():
        cleaned = [answer.split(": ", 1)[1] for answer in payload["answers"]]
        from_to_scale = any("-" in text for text in cleaned)
        all_cleaned_answers.append(
            {"question": qid, "answer": cleaned, "from_to_scale": from_to_scale}
        )

    return personas, questionnaire, all_cleaned_answers


def _create_llm_prompt_for_persona(persona_row, questionnaire, all_cleaned_answers):
    system_prompt = "Nehme die Perspektive der folgenden Person ein: {persona}"
    prompt = (
        f"Welche der Antwortmöglichkeiten ist die Reaktion der Person auf folgende Frage: "
        f"{placeholder.PROMPT_QUESTIONS}\n"
        f"{placeholder.PROMPT_OPTIONS}\n"
        f"{placeholder.PROMPT_AUTOMATIC_OUTPUT_INSTRUCTIONS}"
    )

    llm_prompt = LLMPrompt(
        questionnaire_source=questionnaire,
        questionnaire_name=str(persona_row.name),
        system_prompt=system_prompt.format(persona=persona_row["persona"]),
        prompt=prompt,
    )

    answer_options = {}
    for answer_bundle in all_cleaned_answers:
        rgm = response_generation.ChoiceResponseGenerationMethod(
            answer_bundle["answer"],
            output_template="Antworte nur mit der exakten Antwort.",
        )

        answer_options[answer_bundle["question"]] = generate_likert_options(
            n=len(answer_bundle["answer"]),
            answer_texts=answer_bundle["answer"],
            only_from_to_scale=answer_bundle["from_to_scale"],
            list_prompt_template="Antwortmöglichkeiten: {options}",
            scale_prompt_template="Antwortmöglichkeiten: {start} bis {end}",
            response_generation_method=rgm,
        )

    llm_prompt.prepare_prompt(answer_options=answer_options)
    return llm_prompt


def _create_json_llm_prompt_for_persona(persona_row, questionnaire, all_cleaned_answers):
    """Create notebook-like prompts but with JSON reasoning output constraints."""
    system_prompt = "Nehme die Perspektive der folgenden Person ein: {persona}"
    prompt = (
        f"Welche der Antwortmöglichkeiten ist die Reaktion der Person auf folgende Frage: "
        f"{placeholder.PROMPT_QUESTIONS}\n"
        f"{placeholder.PROMPT_OPTIONS}\n"
        f"{placeholder.PROMPT_AUTOMATIC_OUTPUT_INSTRUCTIONS}"
    )

    llm_prompt = LLMPrompt(
        questionnaire_source=questionnaire,
        questionnaire_name=str(persona_row.name),
        system_prompt=system_prompt.format(persona=persona_row["persona"]),
        prompt=prompt,
    )

    answer_options = {}
    for answer_bundle in all_cleaned_answers:
        rgm = response_generation.JSONReasoningResponseGenerationMethod(
            output_template=f"Antworte nur im folgenden JSON format:\n{placeholder.JSON_TEMPLATE}"
        )
        answer_options[answer_bundle["question"]] = generate_likert_options(
            n=len(answer_bundle["answer"]),
            answer_texts=answer_bundle["answer"],
            only_from_to_scale=answer_bundle["from_to_scale"],
            list_prompt_template="Antwortmöglichkeiten: {options}",
            scale_prompt_template="Antwortmöglichkeiten: {start} bis {end}",
            response_generation_method=rgm,
        )

    llm_prompt.prepare_prompt(answer_options=answer_options)
    return llm_prompt


def test_notebook_like_prompt_construction():
    """Persona workflow should build prompts with question options and output instructions."""
    personas, questionnaire, all_cleaned_answers = _build_notebook_like_inputs()
    llm_prompts = personas.apply(
        _create_llm_prompt_for_persona,
        axis=1,
        args=(questionnaire, all_cleaned_answers),
    ).to_list()

    assert len(llm_prompts) == 2

    system_msg, user_prompt = llm_prompts[0].get_prompt_for_questionnaire_type(item_id="mp18")
    assert "Person A" in system_msg
    assert "Wie zufrieden sind Sie mit der Demokratie?" in user_prompt
    assert "Antwortmöglichkeiten:" in user_prompt
    assert "Antworte nur mit der exakten Antwort." in user_prompt


def test_notebook_like_single_item_end_to_end_with_mocked_generation(monkeypatch):
    """Single-item survey should produce merged results for all personas and items."""
    personas, questionnaire, all_cleaned_answers = _build_notebook_like_inputs()
    llm_prompts = personas.apply(
        _create_llm_prompt_for_persona,
        axis=1,
        args=(questionnaire, all_cleaned_answers),
    ).to_list()

    def fake_batch_generation(**kwargs):
        size = len(kwargs["prompts"])
        answers = [f"MOCK_ANSWER_{i}" for i in range(size)]
        return answers, [None] * size, [None] * size

    monkeypatch.setattr(survey_manager, "batch_generation", fake_batch_generation)

    results = survey_manager.conduct_survey_single_item(
        model=object(),
        llm_prompts=llm_prompts,
        print_progress=False,
    )

    parsed = raw_responses(results)
    merged = create_one_dataframe(parsed)

    assert len(results) == 2
    assert merged.shape[0] == 4  # 2 personas * 2 questions
    assert set(merged["questionnaire_name"]) == {"0", "1"}
    assert set(merged["questionnaire_item_id"]) == {"mp18", "pizza"}
    assert all(value.startswith("MOCK_ANSWER_") for value in merged["llm_response"])


def test_notebook_like_sequential_keeps_history_and_respects_prefill(monkeypatch):
    """Sequential flow should preserve prompt history and skip generation for prefilled items."""
    personas, questionnaire, all_cleaned_answers = _build_notebook_like_inputs()
    prompt = personas.apply(
        _create_llm_prompt_for_persona,
        axis=1,
        args=(questionnaire, all_cleaned_answers),
    ).to_list()[0]

    recorded_calls = []

    def fake_batch_turn_by_turn_generation(**kwargs):
        snapshot = {
            "prompts": [list(conv) for conv in kwargs["prompts"]],
            "assistant_messages": [list(conv) for conv in kwargs["assistant_messages"]],
        }
        recorded_calls.append(snapshot)
        batch_size = len(kwargs["system_messages"])
        return (["GEN_0"] * batch_size, [None] * batch_size, [None] * batch_size)

    monkeypatch.setattr(
        survey_manager,
        "batch_turn_by_turn_generation",
        fake_batch_turn_by_turn_generation,
    )

    results = survey_manager.conduct_survey_sequential(
        model=object(),
        llm_prompts=[prompt],
        print_progress=False,
    )
    parsed = raw_responses(results)[prompt]

    assert len(recorded_calls) == 2
    assert len(recorded_calls[0]["prompts"][0]) == 1
    assert recorded_calls[0]["assistant_messages"][0] == []
    assert len(recorded_calls[1]["prompts"][0]) == 2
    assert recorded_calls[1]["assistant_messages"][0] == ["GEN_0"]
    assert parsed["llm_response"].tolist() == ["GEN_0", "GEN_0"]

    prompt_prefilled = prompt.duplicate()
    prompt_prefilled.get_question(1).prefilled_response = "PREFILLED_PIZZA"
    recorded_calls.clear()

    prefilled_results = survey_manager.conduct_survey_sequential(
        model=object(),
        llm_prompts=[prompt_prefilled],
        print_progress=False,
    )
    parsed_prefilled = raw_responses(prefilled_results)[prompt_prefilled]

    assert len(recorded_calls) == 1
    assert parsed_prefilled["llm_response"].tolist() == ["GEN_0", "PREFILLED_PIZZA"]


def test_notebook_like_json_output_parsing_and_merge(monkeypatch):
    """JSON workflow should parse to structured columns and merge across personas."""
    personas, questionnaire, all_cleaned_answers = _build_notebook_like_inputs()
    llm_prompts = personas.apply(
        _create_json_llm_prompt_for_persona,
        axis=1,
        args=(questionnaire, all_cleaned_answers),
    ).to_list()

    def fake_batch_generation(**kwargs):
        size = len(kwargs["prompts"])
        payload = json.dumps({"reasoning": "weil", "answer": "Sehr zufrieden"})
        return [payload] * size, [None] * size, [None] * size

    monkeypatch.setattr(survey_manager, "batch_generation", fake_batch_generation)

    results = survey_manager.conduct_survey_single_item(
        model=object(),
        llm_prompts=llm_prompts,
        print_progress=False,
    )

    parsed = parse_json(results)
    merged = create_one_dataframe(parsed)

    assert merged.shape[0] == 4
    assert set(merged["questionnaire_name"]) == {"0", "1"}
    assert "reasoning" in merged.columns
    assert "answer" in merged.columns
    assert set(merged["reasoning"]) == {"weil"}
    assert set(merged["answer"]) == {"Sehr zufrieden"}


def test_notebook_like_battery_end_to_end_with_mocked_generation(monkeypatch):
    """Battery flow should aggregate all questions into one response per persona."""
    personas, questionnaire, all_cleaned_answers = _build_notebook_like_inputs()
    llm_prompts = personas.apply(
        _create_llm_prompt_for_persona,
        axis=1,
        args=(questionnaire, all_cleaned_answers),
    ).to_list()

    captured_prompts = []

    def fake_batch_generation(**kwargs):
        captured_prompts[:] = kwargs["prompts"]
        size = len(kwargs["prompts"])
        answers = [f"BATTERY_ANSWER_{i}" for i in range(size)]
        return answers, [None] * size, [None] * size

    monkeypatch.setattr(survey_manager, "batch_generation", fake_batch_generation)

    results = survey_manager.conduct_survey_battery(
        model=object(),
        llm_prompts=llm_prompts,
        print_progress=False,
        item_separator=" || ",
    )

    parsed = raw_responses(results)
    merged = create_one_dataframe(parsed)

    assert len(results) == 2
    assert merged.shape[0] == 2  # one battery response per persona
    assert set(merged["questionnaire_name"]) == {"0", "1"}
    assert set(merged["questionnaire_item_id"]) == {-1}
    assert all(value.startswith("BATTERY_ANSWER_") for value in merged["llm_response"])
    assert len(captured_prompts) == 2
    assert "Wie zufrieden sind Sie mit der Demokratie?" in captured_prompts[0]
    assert "Wie stehen Sie zu Ananas auf Pizza?" in captured_prompts[0]
    assert " || " in captured_prompts[0]



import queue
import threading
import time
from contextlib import redirect_stderr

import streamlit as st
from gui_elements.paginator import paginator
from gui_elements.stateful_widget import StatefulWidgets
from openai import AsyncOpenAI

from qstn.inference.response_generation import JSONResponseGenerationMethod
from qstn.parser.llm_answer_parser import parse_json, parse_json_battery, raw_responses
from qstn.survey_manager import (
    conduct_survey_battery,
    conduct_survey_sequential,
    conduct_survey_single_item,
)
from qstn.utilities.constants import QuestionnairePresentation
from qstn.utilities.utils import create_one_dataframe

# Set OpenAI's API key and API base to use vLLM's API server.

st.page_link(
    "pages/00_Tutorial.py",
    label="📖 Need help? See tutorial: Run & results",
    query_params={"section": "overview"},
)

if "questionnaires" not in st.session_state:
    st.error("You need to first upload a questionnaire and the population you want to survey.")
    st.stop()
    disabled = True
else:
    disabled = False


@st.cache_data
def create_stateful_widget() -> StatefulWidgets:
    return StatefulWidgets()


state = create_stateful_widget()

current_index = paginator(st.session_state.questionnaires, "overview_page")

questionnaires = st.session_state.questionnaires[current_index]

col_llm, col_prompt_display = st.columns(2)

with col_llm:
    st.subheader("⚙️ Inference Parameters")

    with st.container(border=True):
        st.subheader("Core Settings")
        model_name = state.create(
            st.text_input,
            "model_name",
            "Model Name",
            disabled=True,
            help="The model to use for the inference call.",
        )

        temperature = state.create(
            st.slider,
            "temperature",
            "Temperature",
            min_value=0.0,
            max_value=2.0,
            step=0.01,
            initial_value=1.0,
            disabled=True,
            help="Controls randomness. Lower values are more deterministic and less creative.",
        )

        max_tokens = state.create(
            st.number_input,
            "max_tokens",
            "Max Tokens",
            initial_value=1024,
            min_value=1,
            disabled=True,
            help="The maximum number of tokens to generate in the completion.",
        )

        top_p = state.create(
            st.slider,
            "top_p",
            "Top P",
            min_value=0.0,
            max_value=1.0,
            step=0.01,
            initial_value=1.0,
            disabled=True,
            help=(
                "Controls nucleus sampling. The model considers tokens with "
                "top_p probability mass."
            ),
        )

        seed = state.create(
            st.number_input,
            "seed",
            "Seed",
            initial_value=42,
            min_value=0,
            disabled=True,
            help="A specific seed for reproducibility of results.",
        )

        with st.expander("Advanced Inference Settings (JSON)"):
            advanced_inference_params_str = state.create(
                st.text_area,
                "advanced_inference_params_str",
                "JSON for other inference parameters",
                initial_value="",
                height=150,
                disabled=True,
                help=(
                    'Enter any other valid inference parameters like "stop", '
                    '"logit_bias", or "frequency_penalty" as a JSON object.'
                ),
            )


with col_prompt_display:
    st.subheader("📄 Live Preview")

    # Survey method selector
    survey_method_options = {
        "Single item": ("single_item", QuestionnairePresentation.SINGLE_ITEM),
        "Battery": ("battery", QuestionnairePresentation.BATTERY),
        "Sequential": ("sequential", QuestionnairePresentation.SEQUENTIAL),
    }

    survey_method_display = state.create(
        st.selectbox,
        "survey_method",
        "Questionnaire Method",
        options=list(survey_method_options.keys()),
        initial_value="Single item",
        help=(
            "Choose how to conduct the questionnaire: Single item "
            "(one at a time), Battery (all questions together), or "
            "Sequential (with conversation history)."
        ),
    )

    # Get the method name and questionnaire type from selection
    selected_method_name, selected_questionnaire_type = survey_method_options[survey_method_display]

    with st.container(border=True):
        # Add a visible questionnaire identifier to force Streamlit to update when switching
        # This ensures the preview updates even when "change all" makes content identical
        st.caption(
            f"📋 Questionnaire {current_index + 1} of " f"{len(st.session_state.questionnaires)}"
        )

        # For single item mode, show multiple previews (up to 3 items)
        if selected_questionnaire_type == QuestionnairePresentation.SINGLE_ITEM:
            num_questions = len(questionnaires._questions)
            num_previews = min(3, num_questions)  # Show up to 3 previews

            if num_previews > 1:
                st.write(f"**Preview of first {num_previews} items:**")
            else:
                st.write("**Preview:**")

            for i in range(num_previews):
                if num_previews > 1:
                    st.write(f"**Item {i+1}:**")

                current_system_prompt, current_prompt = (
                    questionnaires.get_prompt_for_questionnaire_type(
                        selected_questionnaire_type,
                        item_position=i,
                    )
                )
                current_system_prompt = current_system_prompt.replace("\n", "  \n")
                current_prompt = current_prompt.replace("\n", "  \n")
                # Use a unique container key for each questionnaire/item to
                # force Streamlit to update when switching
                # This ensures the preview updates even when content is identical
                with st.container(key=f"preview_container_{current_index}_{i}"):
                    st.markdown(current_system_prompt)
                    st.write(current_prompt)

                # Add separator between items (except for the last one)
                if i < num_previews - 1:
                    st.divider()
        else:
            # For battery and sequential, show single preview as before
            current_system_prompt, current_prompt = (
                questionnaires.get_prompt_for_questionnaire_type(selected_questionnaire_type)
            )
            current_system_prompt = current_system_prompt.replace("\n", "  \n")
            current_prompt = current_prompt.replace("\n", "  \n")
            # Use a unique container key for each questionnaire to force
            # Streamlit to update when switching
            # This ensures the preview updates even when content is identical
            with st.container(key=f"preview_container_{current_index}"):
                st.markdown(current_system_prompt)
                st.write(current_prompt)


if st.button("Confirm and Run Questionnaire", type="primary", use_container_width=True):
    st.write("Starting inference...")

    openai_api_key = "EMPTY"
    openai_api_base = "http://localhost:8000/v1"

    client = AsyncOpenAI(**st.session_state.client_config)

    inference_config = st.session_state.inference_config.copy()

    model_name = inference_config.pop("model")

    progress_text = st.empty()

    log_queue = queue.Queue()
    result_queue = queue.Queue()

    class QueueWriter:
        def __init__(self, q):
            self.q = q

        def write(self, message):
            if message.strip():
                self.q.put(message)

        def flush(self):
            # This function is needed to match the file-like object interface
            # but we don't need to do anything here.
            pass

    # Helper function for asyncronous runs
    def run_async_in_thread(
        result_q, client, questionnaires, model_name, survey_method_name, **inference_config
    ):
        queue_writer = QueueWriter(log_queue)

        # We need to redirect the output to a queue, as streamlit does not support multithreading
        # API concurrency should be  configurable in the GUI
        try:
            with redirect_stderr(queue_writer):
                # Select the appropriate survey method based on user choice
                if survey_method_name == "single_item":
                    survey_func = conduct_survey_single_item
                elif survey_method_name == "battery":
                    survey_func = conduct_survey_battery
                elif survey_method_name == "sequential":
                    survey_func = conduct_survey_sequential
                else:
                    survey_func = conduct_survey_single_item  # Default fallback

                result = survey_func(
                    client,
                    llm_prompts=questionnaires,
                    client_model_name=model_name,
                    api_concurrency=100,
                    **inference_config,
                )

        except Exception as e:
            result = e
            st.error(e)
        finally:
            result_q.put(result)

    while not log_queue.empty():
        log_queue.get()
    while not result_queue.empty():
        result_queue.get()

    # Get the selected survey method
    survey_method_display = st.session_state.get("survey_method", "Single item")
    survey_method_options = {
        "Single item": ("single_item", QuestionnairePresentation.SINGLE_ITEM),
        "Battery": ("battery", QuestionnairePresentation.BATTERY),
        "Sequential": ("sequential", QuestionnairePresentation.SEQUENTIAL),
    }
    selected_method_name, _ = survey_method_options.get(
        survey_method_display,
        ("single_item", QuestionnairePresentation.SINGLE_ITEM),
    )

    thread = threading.Thread(
        target=run_async_in_thread,
        args=(
            result_queue,
            client,
            st.session_state.questionnaires,
            model_name,
            selected_method_name,
        ),
        kwargs=inference_config,
    )
    thread.start()

    all_questions_placeholder = st.empty()
    progress_placeholder = st.empty()

    while thread.is_alive():
        try:
            # Here we can write directly to the UI, as it is the main thread
            # TQDM uses carriage returns (\r) to animate in the console, we only show clear lines
            log_message = log_queue.get_nowait()
            # This is quite a hacky solution for now, we should adjust QSTN
            # to make the messages clearly parsable.
            if "[A" not in log_message and "Processing Prompts" not in log_message:
                all_questions_placeholder.text(log_message.strip().replace("\r", ""))

            elif "Processing Prompts" in log_message:
                progress_placeholder.text(log_message.strip().replace("\r", ""))

        except queue.Empty:
            pass
        time.sleep(0.1)
    thread.join()

    all_questions_placeholder.empty()
    progress_placeholder.empty()

    try:
        final_output = result_queue.get_nowait()
    except queue.Empty:
        st.error("Could not retrieve result from the asynchronous task.")

    st.success("Finished inferencing!")

    # Check if any questionnaire uses JSON response generation methods
    has_json_rgm = False
    if isinstance(final_output, list) and len(final_output) > 0:
        for result in final_output:
            if hasattr(result, "questionnaire") and hasattr(result.questionnaire, "_questions"):
                for question in result.questionnaire._questions:
                    if (
                        hasattr(question, "answer_options")
                        and question.answer_options
                        and hasattr(question.answer_options, "response_generation_method")
                        and question.answer_options.response_generation_method
                    ):
                        rgm = question.answer_options.response_generation_method
                        if isinstance(rgm, JSONResponseGenerationMethod):
                            has_json_rgm = True
                            break
                if has_json_rgm:
                    break

    # Use appropriate parser based on response generation method
    if has_json_rgm:
        # Check survey method to use correct parser
        survey_method_display = st.session_state.get("survey_method", "Single item")
        if survey_method_display == "Battery":
            responses = parse_json_battery(final_output)
        else:
            responses = parse_json(final_output)
    else:
        responses = raw_responses(final_output)

    df = create_one_dataframe(responses)

    # Store the dataframe in session state for saving later
    st.session_state.results_dataframe = df
    st.session_state.inference_completed = True

    st.dataframe(df)

# Show save button if inference is completed
if "inference_completed" in st.session_state and st.session_state.inference_completed:
    st.divider()
    st.subheader("💾 Save Results")

    # Text input for filename
    if "save_filename" not in st.session_state:
        st.session_state.save_filename = "questionnaire_results.csv"

    save_filename = st.text_input(
        "Save File",
        value=st.session_state.save_filename,
        key="save_filename_input",
        help=("Enter the filename for the results. Should be a CSV file " "(e.g., results.csv)."),
    )

    # Ensure filename ends with .csv
    if save_filename and not save_filename.endswith(".csv"):
        save_filename = save_filename + ".csv"

    # Convert dataframe to CSV string for download
    csv = st.session_state.results_dataframe.to_csv(index=False)

    st.download_button(
        label="Save Results",
        data=csv,
        file_name=save_filename if save_filename else "questionnaire_results.csv",
        mime="text/csv",
        type="primary",
        use_container_width=True,
        help=(
            "Click to save the results to your computer. You can choose "
            "the directory and filename in the save dialog."
        ),
    )



"""Tests for the top-level persona creator module."""

import pandas as pd
import pytest

from qstn import survey_manager
from qstn.parser.llm_answer_parser import raw_responses
from qstn.persona_creator import LLMConfig, PersonaBuilder
from qstn.prompt_builder import LLMPrompt


def test_render_automatic_supports_order_templates_and_missing_policy():
    builder = PersonaBuilder()
    attributes = {"age": 30, "city": "Berlin", "income": None}

    text_skip = builder.render_automatic(
        attributes,
        attribute_order=["city", "income", "age"],
        attribute_templates={"city": "City={value}", "age": "Age={value}"},
        missing_value_policy="skip",
    )
    assert text_skip == "City=Berlin\nAge=30"

    text_placeholder = builder.render_automatic(
        attributes,
        attribute_order=["city", "income", "age"],
        missing_value_policy="placeholder",
        missing_value_placeholder="n/a",
    )
    assert text_placeholder == "city: Berlin\nincome: n/a\nage: 30"


def test_render_automatic_normalizes_integer_like_floats_and_booleans():
    builder = PersonaBuilder()
    text = builder.render_automatic({"age": 19.0, "is_student": True})
    assert "age: 19" in text
    assert "is_student: Yes" in text


def test_render_automatic_supports_custom_bool_text_and_float_cast_override():
    builder = PersonaBuilder()
    text = builder.render_automatic(
        {"age": 19.0, "is_student": False},
        true_text="JA",
        false_text="NEIN",
        cast_integer_like_floats=False,
    )
    assert "age: 19.0" in text
    assert "is_student: NEIN" in text


def test_attribute_aliases_support_renaming_without_dataframe_preprocessing():
    builder = PersonaBuilder()
    text = builder.render_automatic(
        {"years_experience": 19.0},
        attribute_aliases={"years_experience": "years_of_experience"},
    )
    assert text == "years_of_experience: 19"


def test_render_interview_text_supports_default_and_custom_questions():
    builder = PersonaBuilder(default_interview_question_map={"age": "How old are you?"})
    text = builder.render_interview_text(
        {"age": 30, "city": "Berlin"},
        attribute_order=["age", "city"],
        question_template="What about your {attribute}?",
    )

    assert "Q: How old are you?" in text
    assert "A: 30" in text
    assert "Q: What about your city?" in text
    assert "A: Berlin" in text


def test_build_persona_dataframe_supports_method_specific_kwargs_bundles():
    builder = PersonaBuilder()
    personas = pd.DataFrame([{"years_experience": 19.0, "is_manager": True, "city": "Berlin"}])
    result = builder.build_persona_dataframe(
        personas,
        attribute_columns=["years_experience", "is_manager", "city"],
        automatic_kwargs={
            "attribute_aliases": {"years_experience": "years_of_experience"},
            "attribute_templates": {
                "years_of_experience": "Experience: {value}",
                "is_manager": "Manager: {value}",
            },
        },
        interview_kwargs={
            "attribute_aliases": {"years_experience": "years_of_experience"},
            "question_map": {"years_of_experience": "How many years of experience do you have?"},
        },
        human_prompt_kwargs={
            "attribute_aliases": {"years_experience": "years_of_experience"},
        },
    )
    assert "Experience: 19" in result["persona_automatic"].iloc[0]
    assert "Manager: Yes" in result["persona_automatic"].iloc[0]
    assert "How many years of experience do you have?" in result["persona_interview"].iloc[0]
    assert "- years_of_experience: 19" in result["persona_human_prompt"].iloc[0]


def test_build_human_prompt_and_generate_human_description(monkeypatch):
    builder = PersonaBuilder()
    attrs = {"age": 30, "city": "Berlin"}

    prompt = builder.build_human_prompt(attrs, attribute_order=["city", "age"])
    assert "- city: Berlin" in prompt
    assert "- age: 30" in prompt

    captured = {}

    def fake_batch_generation(**kwargs):
        captured["prompts"] = kwargs["prompts"]
        captured["system_messages"] = kwargs["system_messages"]
        return (["Generated persona"], [None], [None])

    monkeypatch.setattr(
        "qstn.persona_creator.persona_creator.batch_generation", fake_batch_generation
    )
    output = builder.generate_human_description(
        model=object(),
        attributes=attrs,
        attribute_order=["city", "age"],
        llm_config=LLMConfig(print_progress=False),
    )

    assert output == "Generated persona"
    assert len(captured["prompts"]) == 1
    assert "- city: Berlin" in captured["prompts"][0]


def test_build_and_generate_interview_from_automatic(monkeypatch):
    builder = PersonaBuilder()
    automatic_text = "age: 30\ncity: Berlin"
    prompt = builder.build_interview_from_automatic_prompt(automatic_text)
    assert automatic_text in prompt
    assert "Keep each answer value exactly as provided" in prompt

    captured = {}

    def fake_batch_generation(**kwargs):
        captured["prompts"] = kwargs["prompts"]
        return (["Q: How old are you?\nA: 30"], [None], [None])

    monkeypatch.setattr(
        "qstn.persona_creator.persona_creator.batch_generation", fake_batch_generation
    )
    output = builder.generate_interview_from_automatic(
        model=object(),
        automatic_persona=automatic_text,
        llm_config=LLMConfig(print_progress=False),
    )
    assert output.startswith("Q:")
    assert automatic_text in captured["prompts"][0]


def test_build_persona_dataframe_creates_mode_columns_and_default_system_prompt():
    builder = PersonaBuilder()
    personas = pd.DataFrame(
        [
            {"id": 1, "age": 30, "city": "Berlin"},
            {"id": 2, "age": 40, "city": "Hamburg"},
        ]
    )

    result = builder.build_persona_dataframe(
        personas,
        attribute_columns=["age", "city"],
        attribute_order=["city", "age"],
    )

    assert "persona_automatic" in result.columns
    assert "persona_interview" in result.columns
    assert "persona_human_prompt" in result.columns
    assert "persona_human_llm" not in result.columns
    assert "system_prompt" in result.columns
    assert result["system_prompt"].tolist() == result["persona_automatic"].tolist()


def test_build_persona_dataframe_optional_human_llm_mode(monkeypatch):
    builder = PersonaBuilder()
    personas = pd.DataFrame([{"age": 30, "city": "Berlin"}])

    def fake_batch_generation(**kwargs):
        return (["Human profile"], [None], [None])

    monkeypatch.setattr(
        "qstn.persona_creator.persona_creator.batch_generation", fake_batch_generation
    )
    result = builder.build_persona_dataframe(
        personas,
        attribute_columns=["age", "city"],
        include_human_llm=True,
        model=object(),
        default_system_prompt_mode="human_llm",
        llm_config=LLMConfig(print_progress=False),
    )
    assert result["persona_human_llm"].tolist() == ["Human profile"]
    assert result["system_prompt"].tolist() == ["Human profile"]

    with pytest.raises(ValueError, match="`model` must be provided"):
        builder.build_persona_dataframe(personas, include_human_llm=True, model=None)


def test_build_persona_dataframe_optional_interview_llm_mode(monkeypatch):
    builder = PersonaBuilder()
    personas = pd.DataFrame([{"age": 30, "city": "Berlin"}])

    captured = {}

    def fake_batch_generation(**kwargs):
        captured["prompts"] = kwargs["prompts"]
        return (["Q: Age?\nA: 30"], [None], [None])

    monkeypatch.setattr(
        "qstn.persona_creator.persona_creator.batch_generation", fake_batch_generation
    )
    result = builder.build_persona_dataframe(
        personas,
        attribute_columns=["age", "city"],
        include_interview_llm=True,
        model=object(),
        default_system_prompt_mode="interview_llm",
        llm_config=LLMConfig(print_progress=False),
    )
    assert result["persona_interview_llm"].tolist() == ["Q: Age?\nA: 30"]
    assert result["system_prompt"].tolist() == ["Q: Age?\nA: 30"]
    assert "age: 30" in captured["prompts"][0]

    with pytest.raises(ValueError, match="`model` must be provided"):
        builder.build_persona_dataframe(personas, include_interview_llm=True, model=None)


def test_build_persona_dataframe_reuses_interview_questions_once(monkeypatch):
    builder = PersonaBuilder()
    personas = pd.DataFrame(
        [
            {"age": 30, "city": "Berlin"},
            {"age": 40, "city": "Hamburg"},
        ]
    )
    counter = {"calls": 0}
    original = builder._resolve_interview_question_templates

    def wrapped(*args, **kwargs):
        counter["calls"] += 1
        return original(*args, **kwargs)

    monkeypatch.setattr(builder, "_resolve_interview_question_templates", wrapped)
    result = builder.build_persona_dataframe(
        personas,
        attribute_columns=["age", "city"],
        reuse_interview_questions=True,
    )

    assert counter["calls"] == 1
    assert "Q: What is your age?" in result["persona_interview"].iloc[0]
    assert "A: 30" in result["persona_interview"].iloc[0]
    assert "A: 40" in result["persona_interview"].iloc[1]


def test_reused_interview_questions_still_allow_value_in_question_template():
    builder = PersonaBuilder()
    personas = pd.DataFrame(
        [
            {"age": 30, "city": "Berlin"},
            {"age": 40, "city": "Hamburg"},
        ]
    )
    result = builder.build_persona_dataframe(
        personas,
        attribute_columns=["age", "city"],
        interview_kwargs={"question_template": "Is your {attribute} {value}?"},
        reuse_interview_questions=True,
    )
    assert "Q: Is your age 30?" in result["persona_interview"].iloc[0]
    assert "Q: Is your age 40?" in result["persona_interview"].iloc[1]


def test_apply_interview_as_prefilled_turns_prepends_and_keeps_order():
    questionnaire = pd.DataFrame(
        [
            {"questionnaire_item_id": 1, "question_content": "Q1"},
            {"questionnaire_item_id": 2, "question_content": "Q2"},
        ]
    )
    prompt = LLMPrompt(questionnaire_source=questionnaire)
    builder = PersonaBuilder()

    transformed = builder.apply_interview_as_prefilled_turns(
        prompt,
        attributes={"age": 30, "city": "Berlin"},
        attribute_order=["age", "city"],
    )

    assert len(prompt) == 2
    assert len(transformed) == 4

    transformed_questions = transformed.get_questions()
    assert str(transformed_questions[0].item_id).startswith("__persona_interview__")
    assert transformed_questions[0].prefilled_response == "30"
    assert transformed_questions[1].prefilled_response == "Berlin"
    assert transformed_questions[2].item_id == 1
    assert transformed_questions[3].item_id == 2


def test_prefilled_interview_turns_seed_sequential_flow(monkeypatch):
    questionnaire = pd.DataFrame([{"questionnaire_item_id": 1, "question_content": "Final Q?"}])
    prompt = LLMPrompt(questionnaire_source=questionnaire, questionnaire_name="seeded")
    builder = PersonaBuilder()
    transformed = builder.apply_interview_as_prefilled_turns(
        prompt,
        attributes={"age": 30, "city": "Berlin"},
        attribute_order=["age", "city"],
    )

    recorded_calls = []

    def fake_batch_turn_by_turn_generation(**kwargs):
        recorded_calls.append(
            {
                "prompts": [list(conv) for conv in kwargs["prompts"]],
                "assistant_messages": [list(conv) for conv in kwargs["assistant_messages"]],
            }
        )
        return (["GENERATED_FINAL"], [None], [None])

    monkeypatch.setattr(
        survey_manager, "batch_turn_by_turn_generation", fake_batch_turn_by_turn_generation
    )

    results = survey_manager.conduct_survey_sequential(
        model=object(),
        llm_prompts=[transformed],
        print_progress=False,
    )
    parsed = raw_responses(results)[transformed]

    assert len(recorded_calls) == 1
    assert len(recorded_calls[0]["prompts"][0]) == 3
    assert recorded_calls[0]["assistant_messages"][0] == ["30", "Berlin"]
    assert (
        parsed.loc[parsed["questionnaire_item_id"] == 1, "llm_response"].item() == "GENERATED_FINAL"
    )



"""Tests for package and docs version consistency."""

import subprocess
from importlib import util as importlib_util
from pathlib import Path

import qstn


def test_qstn_exposes_non_empty_version() -> None:
    assert hasattr(qstn, "__version__")
    assert isinstance(qstn.__version__, str)
    assert qstn.__version__


def test_docs_release_matches_runtime_version() -> None:
    conf_path = Path(__file__).resolve().parents[1] / "docs" / "conf.py"
    spec = importlib_util.spec_from_file_location("qstn_docs_conf", conf_path)
    assert spec is not None
    assert spec.loader is not None
    docs_conf = importlib_util.module_from_spec(spec)
    spec.loader.exec_module(docs_conf)

    repo_root = Path(__file__).resolve().parents[1]
    try:
        output = subprocess.check_output(
            ["git", "tag", "--sort=-v:refname"],
            cwd=repo_root,
            text=True,
        )
    except (subprocess.CalledProcessError, FileNotFoundError):
        output = ""

    latest_stable_tag = None
    for raw_tag in output.splitlines():
        tag = raw_tag.strip()
        if not tag:
            continue
        normalized = tag[1:] if tag.startswith("v") else tag
        parts = normalized.split(".")
        if len(parts) == 3 and all(part.isdigit() for part in parts):
            latest_stable_tag = normalized
            break

    expected_release = latest_stable_tag or qstn.__version__
    assert docs_conf.release == expected_release



import asyncio
import atexit
import threading
import weakref
from concurrent.futures import Future
from typing import Any

from openai import AsyncOpenAI
from tqdm.asyncio import tqdm_asyncio

from ..utilities.utils import _make_cache_key, generate_seeds
from .dynamic_pydantic import build_pydantic_model_from_json_object
from .reasoning_parser import parse_reasoning
from .response_generation import (
    ChoiceResponseGenerationMethod,
    JSONResponseGenerationMethod,
    LogprobResponseGenerationMethod,
    ResponseGenerationMethod,
)
from .utils import normalize_system_messages


class _ClientLoopRunner:
    """Persistent event loop running in a background thread."""

    def __init__(self):
        self._loop: asyncio.AbstractEventLoop | None = None
        self._loop_ready = threading.Event()

        self._thread = threading.Thread(target=self._thread_main, daemon=True)
        self._thread.start()

        self._loop_ready.wait()

    def _thread_main(self) -> None:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

        self._loop = loop
        self._loop_ready.set()

        loop.run_forever()

        pending = asyncio.all_tasks(loop=loop)

        for task in pending:
            task.cancel()

        if pending:
            loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))

        loop.close()

    def submit(self, coro: Any) -> Future:
        if self._loop is None:
            raise RuntimeError("Loop runner is not initialized.")

        return asyncio.run_coroutine_threadsafe(coro, self._loop)

    def shutdown(self) -> None:
        if self._loop is None:
            return

        if self._loop.is_closed():
            return

        self._loop.call_soon_threadsafe(self._loop.stop)

        if threading.current_thread() is not self._thread:
            self._thread.join(timeout=5)


_CLIENT_LOOP_RUNNERS: weakref.WeakKeyDictionary[AsyncOpenAI, _ClientLoopRunner] = (
    weakref.WeakKeyDictionary()
)
_CLIENT_LOOP_RUNNERS_LOCK = threading.Lock()


def _get_or_create_runner(client: AsyncOpenAI) -> _ClientLoopRunner:
    with _CLIENT_LOOP_RUNNERS_LOCK:
        runner = _CLIENT_LOOP_RUNNERS.get(client)

        if runner is None:
            runner = _ClientLoopRunner()
            _CLIENT_LOOP_RUNNERS[client] = runner

        return runner


def _shutdown_all_client_loop_runners() -> None:
    with _CLIENT_LOOP_RUNNERS_LOCK:
        runners = list(_CLIENT_LOOP_RUNNERS.values())
        _CLIENT_LOOP_RUNNERS.clear()

    for runner in runners:
        runner.shutdown()


atexit.register(_shutdown_all_client_loop_runners)


def run_openai_batch(
    model: AsyncOpenAI,
    system_messages: list[str | None] | None = ("You are a helpful assistant.",),
    prompts: list[str] = ("Hi there! What is your name?",),
    response_generation_method: (
        ResponseGenerationMethod | list[ResponseGenerationMethod] | None
    ) = None,
    seed: int = 42,
    client_model_name: str | None = None,
    api_concurrency: int = 10,
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    print_progress: bool = True,
    **generation_kwargs: Any,
) -> tuple[list[str], list[str], list[str]]:
    normalized_system_messages = normalize_system_messages(
        system_messages=system_messages,
        batch_size=len(prompts),
    )

    # Prepare batch of messages
    batch_messages: list[list[dict[str, str]]] = []
    for system_message, prompt in zip(normalized_system_messages, prompts):
        messages: list[dict[str, str]] = [{"role": "user", "content": prompt}]
        if system_message is not None:
            messages.insert(0, {"role": "system", "content": system_message})
        batch_messages.append(messages)

    batch_size: int = len(batch_messages)

    seeds = generate_seeds(seed, batch_size=batch_size)

    plain_results, logprob_result, reasoning_output = _run_async_in_thread(
        client=model,
        client_model_name=client_model_name,
        batch_messages=batch_messages,
        seeds=seeds,
        concurrency_limit=api_concurrency,
        response_generation_method=response_generation_method,
        print_progress=print_progress,
        reasoning_start_token=reasoning_start_token,
        reasoning_end_token=reasoning_end_token,
        **generation_kwargs,
    )

    return (plain_results, logprob_result, reasoning_output)


def run_openai_batch_conversation(
    model: AsyncOpenAI,
    system_messages: list[str | None] | None = ("You are a helpful assistant.",),
    prompts: list[list[str]] = (("Hi there! What is your name?",),),
    assistant_messages: list[list[str]] | None = None,
    response_generation_method: (
        ResponseGenerationMethod | list[ResponseGenerationMethod] | None
    ) = None,
    seed: int = 42,
    client_model_name: str | None = None,
    api_concurrency: int = 10,
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    print_progress: bool = True,
    **generation_kwargs: Any,
) -> tuple[list[str], list[str], list[str]]:
    normalized_system_messages = normalize_system_messages(
        system_messages=system_messages,
        batch_size=len(prompts),
    )

    batch_messages = []
    batch_size = len(normalized_system_messages)
    if assistant_messages is None:
        assistant_messages = [[] for _ in range(batch_size)]
    for i in range(batch_size):
        messages = []

        # Add system message
        if normalized_system_messages[i] is not None:
            messages.append({"role": "system", "content": normalized_system_messages[i]})

        num_user_msgs = len(prompts[i])
        num_assistant_msgs = len(assistant_messages[i])

        for j in range(num_user_msgs):
            messages.append({"role": "user", "content": prompts[i][j]})
            if j < num_assistant_msgs:
                messages.append({"role": "assistant", "content": assistant_messages[i][j]})

        batch_messages.append(messages)

    seeds = generate_seeds(seed, batch_size=batch_size)

    plain_results, logprob_result, reasoning_output = _run_async_in_thread(
        client=model,
        client_model_name=client_model_name,
        batch_messages=batch_messages,
        seeds=seeds,
        concurrency_limit=api_concurrency,
        response_generation_method=response_generation_method,
        print_progress=print_progress,
        reasoning_start_token=reasoning_start_token,
        reasoning_end_token=reasoning_end_token,
        **generation_kwargs,
    )

    return (plain_results, logprob_result, reasoning_output)


def _run_async_in_thread(
    client: AsyncOpenAI,
    client_model_name: str | None,
    batch_messages: list[list[dict[str, str]]],
    seeds: list[int],
    concurrency_limit: int = 10,
    print_progress: bool = True,
    response_generation_method: (
        ResponseGenerationMethod | list[ResponseGenerationMethod] | None
    ) = None,
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    **generation_kwargs,
):
    logprob_config = _update_logprob_kwargs(response_generation_method, generation_kwargs)

    sampling_params = _create_structured_output(
        batch_size=len(batch_messages),
        response_generation_method=response_generation_method,
    )
    coro = _run_api_batch_async(
        client=client,
        client_model_name=client_model_name,
        batch_messages=batch_messages,
        seeds=seeds,
        concurrency_limit=concurrency_limit,
        print_progress=print_progress,
        response_generation_method=response_generation_method,
        sampling_params=sampling_params,
        logprob_config=logprob_config,
        reasoning_start_token=reasoning_start_token,
        reasoning_end_token=reasoning_end_token,
        **generation_kwargs,
    )

    runner = _get_or_create_runner(client)
    try:
        future = runner.submit(coro)
    except Exception:
        coro.close()
        raise
    return future.result()


async def _run_api_batch_async(
    client: AsyncOpenAI,
    client_model_name: str,
    batch_messages: list[list[dict[str, str]]],
    seeds: list[int],
    concurrency_limit: int = 10,
    print_progress: bool = True,
    sampling_params: list[dict[str, Any]] = (),
    response_generation_method: (
        ResponseGenerationMethod | list[ResponseGenerationMethod] | None
    ) = None,
    logprob_config: LogprobResponseGenerationMethod | None = None,
    reasoning_start_token: str = "<think>",
    reasoning_end_token: str = "</think>",
    **generation_kwargs,
) -> list[str]:
    semaphore = asyncio.Semaphore(concurrency_limit)

    async def get_completion(
        messages: list,
        seed: int,
        sampling_params: dict[str, Any] | list[str] | None = None,
        response_generation_method: ResponseGenerationMethod | None = None,
        **generation_kwargs,
    ):
        async with semaphore:
            request_kwargs = {
                "model": client_model_name,
                "messages": messages,
                "seed": seed,
                **generation_kwargs,
            }

            if response_generation_method:
                if isinstance(response_generation_method, JSONResponseGenerationMethod):
                    request_kwargs["response_format"] = {
                        "type": "json_schema",
                        "json_schema": {
                            "name": "json_schema",
                            "schema": sampling_params,
                        },
                    }

                elif isinstance(response_generation_method, ChoiceResponseGenerationMethod):
                    import warnings

                    warnings.warn(
                        "Strict Choice Response Generation is only supported for " "vllm APIs.",
                        stacklevel=2,
                    )
                    if True:  # We could use this if we can ensure that the api is vllm.
                        request_kwargs["extra_body"] = {
                            "structured_outputs": {"choice": sampling_params}
                        }
                    else:
                        request_kwargs["response_format"] = {
                            "type": "json_schema",
                            "json_schema": {
                                "name": "Choice",
                                "strict": True,
                                "schema": {
                                    "type": "object",
                                    "properties": {
                                        "selection": {
                                            "type": "string",
                                            "enum": sampling_params,
                                        }
                                    },
                                    "required": ["selection"],
                                    "additionalProperties": False,
                                },
                            },
                        }

            return await client.chat.completions.create(**request_kwargs)

    # pbar = tqdm.tqdm if print_progress else lambda x: x

    if sampling_params:
        tasks = [
            get_completion(messages, seed, struct_output, rgm, **generation_kwargs)
            for messages, seed, struct_output, rgm in zip(
                batch_messages,
                seeds,
                sampling_params,
                response_generation_method,
            )
        ]
    else:
        tasks = [
            get_completion(messages, seed, **generation_kwargs)
            for messages, seed in zip(batch_messages, seeds)
        ]
    if print_progress:
        responses = await tqdm_asyncio.gather(*tasks, total=len(tasks), desc="Processing Prompts")
    else:
        responses = await asyncio.gather(*tasks, return_exceptions=True)

    final_results = []
    reasoning_output = []
    logprob_result = []

    patterns = [
        (reasoning_start_token, reasoning_end_token),
    ]

    for response in responses:
        if isinstance(response, Exception):
            print(f"A request failed permanently after all retries: {response}")
            final_results.append(f"Error: {response}")
        else:
            msg = response.choices[0].message

            # Automatic reasoning parsing
            reasoning = getattr(msg, "reasoning", None) or getattr(msg, "reasoning_content", None)

            if reasoning is None:
                # Fallback to parsing manually
                final_answer, extracted_reasoning = parse_reasoning(msg.content, patterns=patterns)
                final_results.append(final_answer)
                if extracted_reasoning:
                    reasoning_output.append(extracted_reasoning.strip())
                else:
                    reasoning_output.append(extracted_reasoning)
            else:
                final_results.append(msg.content)
                reasoning_output.append(reasoning)
            if logprob_config and response.choices[0].logprobs:
                logprob_result.append(
                    [
                        [{"token": top.token, "logprob": top.logprob} for top in lp.top_logprobs]
                        for lp in response.choices[0].logprobs.content
                    ]
                )
            else:
                logprob_result.append(None)

    return final_results, logprob_result, reasoning_output


def _update_logprob_kwargs(response_generation_method, generation_kwargs):
    logprob_config = None

    if isinstance(response_generation_method, LogprobResponseGenerationMethod):
        logprob_config = response_generation_method
    elif isinstance(response_generation_method, list):
        logprob_config = next(
            (
                item
                for item in response_generation_method
                if isinstance(item, LogprobResponseGenerationMethod)
            ),
            None,
        )
    if logprob_config:
        generation_kwargs["logprobs"] = True
        generation_kwargs["top_logprobs"] = logprob_config.top_logprobs
        if logprob_config.token_limit is not None:
            generation_kwargs["max_tokens"] = logprob_config.token_limit

    return logprob_config


def _create_structured_output(
    batch_size: int,
    response_generation_method: ResponseGenerationMethod | list[ResponseGenerationMethod] | None,
) -> dict[str, Any]:
    """
    Create sampling parameters for generation.

    Args:
        batch_size: Number of prompts in batch
        seeds: Random seeds for generation
        answer_production_method: Output structure configuration
        use_vllm: If True, creates vLLM parameters
        **generation_kwargs: Additional sampling parameters

    Returns:
        Sampling parameters for vLLM or API configuration
    """

    use_structured: bool = response_generation_method and isinstance(
        response_generation_method, (list, ResponseGenerationMethod)
    )

    if use_structured:
        return _create_structured_params(
            batch_size=batch_size,
            response_generation_method=response_generation_method,
        )

    return None


def _create_structured_params(
    batch_size: int,
    response_generation_method: ResponseGenerationMethod | list[ResponseGenerationMethod],
) -> list[dict[str, Any]]:

    structured_output = []

    # Same for all calls
    if isinstance(response_generation_method, ResponseGenerationMethod):
        if isinstance(response_generation_method, JSONResponseGenerationMethod):
            pydantic_model = build_pydantic_model_from_json_object(
                json_object=response_generation_method.json_object,
            )
            json_schema = pydantic_model.model_json_schema()
            structured_output = [json_schema] * batch_size
        elif (
            isinstance(
                response_generation_method,
                (ChoiceResponseGenerationMethod, LogprobResponseGenerationMethod),
            )
            and response_generation_method.allowed_choices is not None
        ):
            _allowed_choices = [str(c) for c in response_generation_method.allowed_choices]
            structured_output = [_allowed_choices] * batch_size

    # Different response generation methods for each question
    else:
        structured_output = []
        cache: dict[str, Any] = {}
        for i in range(batch_size):
            current_method = response_generation_method[i]
            if isinstance(current_method, JSONResponseGenerationMethod):
                key = _make_cache_key(current_method.get_json_prompt(), None)

                if key not in cache:
                    pydantic_model = build_pydantic_model_from_json_object(
                        json_object=current_method.json_object,
                    )
                    json_schema = pydantic_model.model_json_schema()
                    cache[key] = json_schema

                structured_output.append(cache[key])
            elif (
                isinstance(
                    current_method,
                    (ChoiceResponseGenerationMethod, LogprobResponseGenerationMethod),
                )
                and current_method.allowed_choices is not None
            ):
                _allowed_choices = [str(c) for c in current_method.allowed_choices]

                key = _make_cache_key(_allowed_choices, None)
                if key not in cache:
                    cache[key] = _allowed_choices
                structured_output.append(cache[key])
            else:
                structured_output.append(None)

    return structured_output



"""Smoke tests for constants and prompt template primitives."""

from qstn.utilities import constants, placeholder, prompt_templates


def test_constants_values():
    """Core constant values should be defined and of expected types."""
    assert constants.SYSTEM_PROMPT_FIELD == "system_prompt"
    assert constants.QUESTIONNAIRE_ITEM_ID == "questionnaire_item_id"


def test_templates_contain_placeholders():
    """Templates should format correctly and include placeholder markers."""
    # pick a couple of templates, ensure formatting works
    s = prompt_templates.LIST_OPTIONS_DEFAULT.format(options="x")
    assert "x" in s
    t = prompt_templates.SCALE_OPTIONS_DEFAULT.format(start="1", end="5")
    assert "1" in t and "5" in t
    # placeholder constants are braces strings
    assert placeholder.PROMPT_QUESTIONS.startswith("{{")



"""
Tutorial & Help page for the QSTN GUI.
Use ?section=start|options|prompts|inference|overview to deep-link to a section.
"""

import streamlit as st

st.set_page_config(page_title="Tutorial & Help", layout="wide")
st.title("📖 QSTN Tutorial & Help")
st.markdown(
    "Full documentation of every option in the GUI. Use the sections below "
    "or jump to a step from the links."
)

# Section keys for deep linking
SECTION_START = "start"
SECTION_OPTIONS = "options"
SECTION_PROMPTS = "prompts"
SECTION_INFERENCE = "inference"
SECTION_OVERVIEW = "overview"

section_param = st.query_params.get("section", SECTION_START)

# Quick jump links
st.markdown("**Jump to:** ")
cols = st.columns(5)
with cols[0]:
    st.page_link(
        "pages/00_Tutorial.py",
        label="Start",
        query_params={"section": SECTION_START},
    )
with cols[1]:
    st.page_link(
        "pages/00_Tutorial.py",
        label="Answer options",
        query_params={"section": SECTION_OPTIONS},
    )
with cols[2]:
    st.page_link(
        "pages/00_Tutorial.py",
        label="Prompts",
        query_params={"section": SECTION_PROMPTS},
    )
with cols[3]:
    st.page_link(
        "pages/00_Tutorial.py",
        label="Inference",
        query_params={"section": SECTION_INFERENCE},
    )
with cols[4]:
    st.page_link(
        "pages/00_Tutorial.py",
        label="Run & results",
        query_params={"section": SECTION_OVERVIEW},
    )

st.divider()

# =============================================================================
# 1. START PAGE
# =============================================================================
with st.expander(
    "**1. Start page – Sessions, questionnaire & population**",
    expanded=(section_param == SECTION_START),
):
    st.markdown(
        "On the **Start** page you create or load a session, upload your data, "
        "and build the survey. All work is saved per session."
    )
    st.markdown(
        "**Session options** (no replica): Continue Last Session, Start New "
        "Session, Switch to another session (dropdown), Create New, View Saved "
        "Sessions (Load/Delete)."
    )
    st.subheader("Data upload")
    with st.container(border=True):
        _c1, _c2 = st.columns(2)
        with _c1:
            st.text_input(
                "Select a questionnaire to start with",
                value="",
                disabled=True,
                key="tutorial_start_quest",
            )
            st.caption(
                "File upload for the **questionnaire** CSV. One row per "
                "question. Required columns: `questionnaire_item_id`, "
                "`question_content`, `question_stem` (optional)."
            )
        with _c2:
            st.text_input(
                "Select a population to start with",
                value="",
                disabled=True,
                key="tutorial_start_pop",
            )
            st.caption(
                "File upload for the **population** CSV. One row per persona. "
                "Required columns: `questionnaire_name`, `system_prompt`, "
                "`questionnaire_instruction`."
            )
        st.button(
            "Confirm and Prepare Questionnaire",
            disabled=True,
            key="tutorial_start_btn",
        )
        st.caption(
            "Builds the survey and goes to **Answer options**. Use **Clear** "
            "under each upload to reset to example data."
        )
    st.subheader("Next step")
    st.page_link("Start_Page.py", label="→ Go to Start page")

# =============================================================================
# 2. ANSWER OPTIONS (LIKERT SCALE OPTIONS GENERATOR)
# =============================================================================
with st.expander(
    "**2. Answer options – Likert scales & response format**",
    expanded=(section_param == SECTION_OPTIONS),
):
    st.markdown(
        "This page configures how the LLM sees answer options and how the "
        "model outputs its answer (free text, JSON, or constrained choice)."
    )
    with st.container(border=True):
        st.subheader("Main Configuration")
        _col1, _col2, _col3 = st.columns(3)
        with _col1:
            st.number_input(
                "Number of Options (n)",
                value=5,
                min_value=2,
                step=1,
                disabled=True,
                key="tutorial_opt_n",
            )
            st.caption(
                "Total number of choices in the scale (e.g. 5 for a 5-point " "scale). Minimum 2."
            )
        with _col2:
            st.selectbox(
                "Index Type",
                options=["integer", "char_low", "char_up"],
                index=0,
                disabled=True,
                key="tutorial_opt_idx",
            )
            st.caption(
                "How options are labeled: integer (1,2,3…), char_low (a,b,c…), "
                "or char_up (A,B,C…)."
            )
        with _col3:
            st.number_input(
                "Starting Index",
                value=1,
                step=1,
                disabled=True,
                key="tutorial_opt_start",
            )
            st.caption("Number to start counting from (e.g. 1).")
        st.subheader("Ordering and Structure")
        _co1, _co2, _co3, _co4 = st.columns(4)
        with _co1:
            st.checkbox(
                "From-To Scale Only",
                value=False,
                disabled=True,
                key="tutorial_opt_ft",
            )
            st.caption(
                "Only first and last labels shown (e.g. “1 Strongly Disagree "
                "to 5 Strongly Agree”). Requires exactly 2 answer texts."
            )
        with _co2:
            st.checkbox(
                "Random Order",
                value=False,
                disabled=True,
                key="tutorial_opt_rand",
            )
            st.caption("Randomize option order per prompt. Cannot combine with " "Reversed Order.")
        with _co3:
            st.checkbox(
                "Reversed Order",
                value=False,
                disabled=True,
                key="tutorial_opt_rev",
            )
            st.caption("Reverse option order (e.g. 5→1).")
        with _co4:
            st.checkbox(
                "Even Order",
                value=False,
                disabled=True,
                key="tutorial_opt_even",
            )
            st.caption("If odd number of labels, middle option is removed.")
        st.subheader("Answer Texts")
        st.text_area(
            "Enter Answer Texts (one per line)",
            value="Strongly Disagree\nDisagree\nNeutral\nAgree\nStrongly Agree",
            height=120,
            disabled=True,
            key="tutorial_opt_texts",
        )
        st.caption(
            "One label per line. Number of lines must match **Number of "
            "Options** (or 2 if From-To Scale Only)."
        )
        st.subheader("Response Generation Method")
        st.checkbox(
            "Output Indices Only",
            value=False,
            disabled=True,
            key="tutorial_opt_outidx",
        )
        st.caption(
            "If checked, model outputs only indices (e.g. 1, 2) instead of "
            "full text (e.g. 1: Strongly Disagree). Applies to all methods "
            "below."
        )
        st.selectbox(
            "Response Generation Method",
            options=[
                "None",
                "JSON Single Answer",
                "JSON All Options (Probabilities)",
                "JSON with Reasoning",
                "Choice",
            ],
            index=0,
            disabled=True,
            key="tutorial_opt_rgm",
        )
        st.markdown("**Details by option:**")
        st.markdown("""
        - **None** — Model returns free text; no structured output.
          You must parse the response yourself. No extra configuration on this page.
        - **JSON Single Answer** — Model returns a single selected answer in JSON
          (e.g. one chosen option per question). Parsing is automatic; use when
          you want one answer per item.
        - **JSON All Options (Probabilities)** — Model returns a probability
          (or score) for each option. Use when you want a distribution over
          choices rather than a single pick. Parsing expects this structure.
        - **JSON with Reasoning** — Model returns reasoning text plus the chosen
          answer in JSON. Use when you need both explanation and a structured
          answer. Parsing extracts both.
        - **Choice** — Model must pick exactly one value from a fixed list.
          When selected, **Allowed Choices** appears: one choice per line,
          either indices only (e.g. 1, 2, 3) or with labels
          (e.g. 1: Strongly Disagree). **Output Indices Only** controls whether
          the model returns indices or full labels. Parsing uses the same list.
        """)
    st.subheader("Next step")
    st.page_link("pages/01_Option_Prompt.py", label="→ Go to Answer options")

# =============================================================================
# 3. PROMPT CONFIGURATION
# =============================================================================
with st.expander(
    "**3. Prompt configuration – System prompt & instructions**",
    expanded=(section_param == SECTION_PROMPTS),
):
    st.markdown(
        "Edit the system and user prompts the LLM sees. Use the checkboxes to "
        "copy the current prompts to all questionnaires on confirm."
    )
    with st.container(border=True):
        st.subheader("Configuration")
        st.checkbox(
            "On update: change all System Prompts",
            value=True,
            disabled=True,
            key="tutorial_prompt_chg_sys",
        )
        st.caption(
            "When you click Confirm and Prepare Questionnaire, the current "
            "system prompt is applied to **all** questionnaires (missing "
            "placeholders added to others)."
        )
        st.checkbox(
            "On update: change all questionnaire instructions",
            value=True,
            disabled=True,
            key="tutorial_prompt_chg_inst",
        )
        st.caption("Same for the main prompt (instructions); all questionnaires get " "this text.")
        st.text_area(
            "System prompt",
            value="You are a student.",
            height=80,
            disabled=True,
            key="tutorial_prompt_sys",
        )
        st.caption("The system message sent to the model (e.g. role or context).")
        st.text_area(
            "Prompt",
            value="Please answer the following questions.",
            height=80,
            disabled=True,
            key="tutorial_prompt_user",
        )
        st.caption("User-facing instructions given before the questions.")
        st.write("**Insert Placeholder:**")
        _pb1, _pb2, _pb3, _pb4 = st.columns(4)
        with _pb1:
            st.button("Prompt Questions", disabled=True, key="tutorial_prompt_btn_p")
        with _pb2:
            st.button("Prompt Options", disabled=True, key="tutorial_prompt_btn_o")
        with _pb3:
            st.button("Automatic Output", disabled=True, key="tutorial_prompt_btn_a")
        with _pb4:
            st.button("JSON Template", disabled=True, key="tutorial_prompt_btn_j")
        st.caption(
            "Type shortcut (-P, -O, -A, -J) in system or main prompt, then "
            "click to replace with placeholder. **Prompt Questions**: where "
            "question text goes. **Prompt Options**: where answer options go "
            "(configure options first). **Automatic Output**: JSON/output "
            "instructions. **JSON Template**: JSON schema for response."
        )
        st.text_area(
            "Question Stem",
            value="How do you feel about?",
            height=80,
            disabled=True,
            key="tutorial_prompt_stem",
        )
        st.caption(
            "Template for each question. Use **Question Content** placeholder "
            "(-Q) so the actual item (e.g. “Coffee”) is inserted."
        )
        st.checkbox(
            "Randomize the order of items",
            value=False,
            disabled=True,
            key="tutorial_prompt_rand",
        )
        st.caption("If checked, question order is randomized per questionnaire.")
    st.caption(
        "**Live Preview** (on the real page) shows prompts with placeholders "
        "filled; use **Update Preview** to refresh. The **paginator** switches "
        "which questionnaire you edit."
    )
    st.subheader("Next step")
    st.page_link(
        "pages/02_Prompt_Configuration.py",
        label="→ Go to Prompt configuration",
    )

# =============================================================================
# 4. INFERENCE SETTINGS
# =============================================================================
with st.expander(
    "**4. Inference settings – API client & model**",
    expanded=(section_param == SECTION_INFERENCE),
):
    st.markdown(
        "Configure where the LLM is called (API URL, key) and how (model, "
        "temperature, etc.). Values are used on Final overview when you run "
        "the survey."
    )
    _inf_col1, _inf_col2 = st.columns(2)
    with _inf_col1:
        with st.container(border=True):
            st.subheader("1. Client Configuration")
            st.text_input(
                "API Key",
                value="",
                type="password",
                disabled=True,
                key="tutorial_inf_apikey",
            )
            st.caption("Your OpenAI API key (or leave empty for local vLLM). Handled " "securely.")
            st.text_input(
                "Base URL",
                value="",
                placeholder="https://api.openai.com/v1",
                disabled=True,
                key="tutorial_inf_base",
            )
            st.caption(
                "API base URL. For **local vLLM** use `http://localhost:8000/v1`. "
                "Optional: Organization ID, Project ID."
            )
            st.number_input(
                "Timeout (seconds)",
                value=20,
                min_value=1,
                disabled=True,
                key="tutorial_inf_timeout",
            )
            st.caption("Request timeout in seconds.")
            st.number_input(
                "Max Retries",
                value=2,
                min_value=0,
                disabled=True,
                key="tutorial_inf_retries",
            )
            st.caption(
                "Number of retries for failed requests. Advanced: JSON for "
                "default_headers, default_query, etc."
            )
    with _inf_col2:
        with st.container(border=True):
            st.subheader("2. Inference Configuration")
            st.text_input(
                "Model Name",
                value="",
                placeholder="meta-llama/Llama-3.1-70B-Instruct",
                disabled=True,
                key="tutorial_inf_model",
            )
            st.caption(
                "Model ID. Must match your API/vLLM (e.g. "
                "`meta-llama/Llama-3.2-3B-Instruct` for vLLM)."
            )
            st.slider(
                "Temperature",
                0.0,
                2.0,
                1.0,
                0.01,
                disabled=True,
                key="tutorial_inf_temp",
            )
            st.caption("Randomness (0–2). Lower = more deterministic.")
            st.number_input(
                "Max Tokens",
                value=1024,
                min_value=1,
                disabled=True,
                key="tutorial_inf_maxtok",
            )
            st.caption("Maximum tokens to generate per completion.")
            st.slider(
                "Top P",
                0.0,
                1.0,
                1.0,
                0.01,
                disabled=True,
                key="tutorial_inf_topp",
            )
            st.caption(
                "Nucleus sampling: consider tokens with cumulative probability " "up to this value."
            )
            st.number_input(
                "Seed",
                value=42,
                min_value=0,
                disabled=True,
                key="tutorial_inf_seed",
            )
            st.caption(
                "Seed for reproducible sampling. Advanced: JSON for stop, "
                "presence_penalty, frequency_penalty, logit_bias."
            )
    st.subheader("Using local vLLM")
    st.markdown("""
    1. Start the vLLM server, e.g.:
       ```bash
       python -m vllm.entrypoints.openai.api_server \
         --model meta-llama/Llama-3.2-3B-Instruct \
         --max-model-len 8192
       ```
    2. In the GUI: set **Base URL** to `http://localhost:8000/v1`
       and **Model name** to the same model ID
       (e.g. `meta-llama/Llama-3.2-3B-Instruct`).
    3. Click **Generate Configuration & Code** to save;
       then go to **Final overview** to run the survey.
    """)
    st.subheader("Next step")
    st.markdown(
        "Click **Generate Configuration & Code** to save client and inference "
        "config and go to **Final overview**."
    )
    st.page_link("pages/03_Inference_Setting.py", label="→ Go to Inference settings")

# =============================================================================
# 5. FINAL OVERVIEW & RUN
# =============================================================================
with st.expander(
    "**5. Final overview – Run survey & save results**",
    expanded=(section_param == SECTION_OVERVIEW),
):
    st.markdown(
        "Read-only inference summary, prompt preview for the selected method, "
        "then run the survey and save results."
    )
    with st.container(border=True):
        st.subheader("⚙️ Inference Parameters")
        st.text_input(
            "Model Name",
            value="meta-llama/Llama-3.2-3B-Instruct",
            disabled=True,
            key="tutorial_overview_model",
        )
        st.caption(
            "Shown from Inference settings (not editable here). Change on "
            "**Inference settings** page."
        )
        st.slider(
            "Temperature",
            0.0,
            2.0,
            1.0,
            0.01,
            disabled=True,
            key="tutorial_overview_temp",
        )
        st.number_input(
            "Max Tokens",
            value=1024,
            min_value=1,
            disabled=True,
            key="tutorial_overview_maxtok",
        )
        st.selectbox(
            "Questionnaire Method",
            options=["Single item", "Battery", "Sequential"],
            index=0,
            disabled=True,
            key="tutorial_overview_method",
        )
        st.caption(
            "**Single item**: one question per API call. **Battery**: all "
            "questions in one call. **Sequential**: all in one thread with "
            "conversation history. Affects prompts and result parsing."
        )
        st.markdown("**Details by option:**")
        st.markdown("""
        - **Single item** — One question is sent per API call; the model sees
          only that question (and system/instruction context). Preview on the
          page shows the first few items so you can check how each is prompted.
        - **Battery** — All questions are sent in one API call; the model sees
          the full list and responds once. Preview shows one combined prompt.
        - **Sequential** — All questions are sent in one conversation thread
          with history; the model sees prior Q&A. Preview shows one combined
          flow. Parsing respects the conversation structure.
        """)
    st.caption(
        "**Live preview** (on the real page): for Single item, first few items "
        "shown; for Battery/Sequential, one combined preview. **Paginator** "
        "switches questionnaire."
    )
    with st.container(border=True):
        st.subheader("💾 Save Results")
        st.text_input(
            "Save File",
            value="questionnaire_results.csv",
            disabled=True,
            key="tutorial_overview_savefile",
        )
        st.caption("Filename for the CSV. If you omit `.csv`, it is appended.")
        st.button("Save Results", disabled=True, key="tutorial_overview_savebtn")
        st.caption(
            "Downloads the results table as CSV. Appears after you click "
            "**Confirm and Run Questionnaire** and inference finishes."
        )
    st.page_link("pages/04_Final_Overview.py", label="→ Go to Final overview")

st.divider()
st.subheader("Workflow summary")
st.markdown(
    "**Start** (upload data, confirm) → **Answer options** (scale + response "
    "method, confirm) → **Prompt configuration** (prompts + placeholders, "
    "confirm) → **Inference settings** (API + model, generate config) → "
    "**Final overview** (run & save)."
)



from typing import Final

# Prompt Placeholder
PROMPT_QUESTIONS: Final[str] = "{{QUESTION_PLACEHOLDER}}"
PROMPT_OPTIONS: Final[str] = "{{OPTIONS_PLACEHOLDER}}"
PROMPT_AUTOMATIC_OUTPUT_INSTRUCTIONS: Final[str] = "{{AUTOMATIC_OUTPUT_PLACEHOLDER}}"

# Question Placeholder
QUESTION_CONTENT: Final[str] = "{{QUESTION_CONTENT_PLACEHOLDER}}"

# JSON Template
JSON_TEMPLATE: Final[str] = "{{JSON_TEMPLATE}}"

# Compact scale range for JSON explanations
SCALE_RANGE: Final[str] = "{{SCALE_RANGE_PLACEHOLDER}}"



"""Direct unit tests for `parse_reasoning` edge cases."""

import pytest

from qstn.inference.reasoning_parser import parse_reasoning


@pytest.mark.parametrize(
    "full_text,patterns,expected",
    [
        # no tags at all, should return text as answer with no reasoning
        ("Hello world", [("<s>", "</s>")], ("Hello world", None)),
        # start and end present
        ("<think>R</think>Answer", [("<think>", "</think>")], ("Answer", "R")),
        # only end tag
        ("R</think>Answer", [("<think>", "</think>")], ("Answer", "R")),
        # only start tag (cutoff)
        ("<think>R", [("<think>", "</think>")], ("", "R")),
        # multiple patterns: first matched used
        ("<a>foo</a>bar", [("<a>", "</a>"), ("<b>", "</b>")], ("bar", "foo")),
        # pattern list empty
        ("<think>here</think>res", [], ("<think>here</think>res", None)),
    ],
)
def test_parse_reasoning_various(full_text, patterns, expected):
    """Ensure all reasoning parsing cases behave as documented."""
    answer, reasoning = parse_reasoning(full_text, patterns)
    assert (answer, reasoning) == expected


def test_parse_reasoning_edge_cases():
    """Check that when tags appear but no closing sequence occurs we still return correct answer."""
    # For multiple tagged sections, parser removes the first section only.
    answer, reasoning = parse_reasoning(
        "<think>a</think><think>b</think>c", [("<think>", "</think>")]
    )
    assert answer == "<think>b</think>c"
    assert reasoning == "a"

    # tag content empty
    answer, reasoning = parse_reasoning("<think></think>", [("<think>", "</think>")])
    assert answer == "" and reasoning == ""



import random
import re
import string

from qstn.inference.survey_inference import batch_generation


def key_typos(text: str, probability: float = 0.1) -> str:
    """
    Randomly replaces characters with random alphabet letters to simulate typos.
    Args:
        text (str): The input text to perturb.
        probability (float): The probability of replacing each character.
    Returns:
        str: The text with random character replacements based on the given probability.
    """
    if not text:
        return text

    # Get all possible letters (a-z and A-Z)
    alphabet = string.ascii_letters

    text_list = list(text)

    for i, char in enumerate(text_list):
        # We check char.isalpha() so we don't replace spaces or punctuation
        if char.isalpha() and random.random() < probability:
            text_list[i] = random.choice(alphabet)

    return "".join(text_list)


def keyboard_typos(text: str, probability: float = 0.1) -> str:
    """
    Introduces typos based on keyboard proximity.
    Args:
        text (str): The input text to perturb.
        probability (float): The probability of introducing a typo for each character.
    Returns:
        str: The text with keyboard-based typos introduced based on the given probability.
    """
    keyboard_neighbors = {
        "a": "qwsz",
        "b": "vghn",
        "c": "xdfv",
        "d": "ersfcx",
        "e": "wsdr",
        "f": "rtgdvc",
        "g": "tyfhbv",
        "h": "yugjbn",
        "i": "ujko",
        "j": "uikhnm",
        "k": "ijolm",
        "l": "opk",
        "m": "njk",
        "n": "bhjm",
        "o": "iklp",
        "p": "ol",
        "q": "wa",
        "r": "edft",
        "s": "awedz",
        "t": "rfgy",
        "u": "yhji",
        "v": "cfgb",
        "w": "qase",
        "x": "zsdc",
        "y": "tugh",
        "z": "asx",
    }

    if not text:
        return text

    text_list = list(text)
    for i in range(len(text_list)):
        char = text_list[i].lower()
        if char in keyboard_neighbors and random.random() < probability:
            neighbors = keyboard_neighbors[char]
            typo_char = random.choice(neighbors)
            # Preserve original case
            if text_list[i].isupper():
                typo_char = typo_char.upper()
            text_list[i] = typo_char
    return "".join(text_list)


def letter_swaps(text: str, probability: float = 0.1) -> str:
    """
    Randomly swaps adjacent letters in the text.
    Args:
        text (str): The input text to perturb.
        probability (float): The probability of swapping each adjacent letter pair.
    Returns:
        str: The text with adjacent letters swapped based on the given probability.
    """
    if not text:
        return text

    text_list = list(text)
    i = 0
    while i < len(text_list) - 1:
        if random.random() < probability:
            text_list[i], text_list[i + 1] = text_list[i + 1], text_list[i]
            i += 2  # Skip next character to avoid double swapping
        else:
            i += 1
    return "".join(text_list)


def make_synonyms(all_prompts: list[str], model: str, instruction: str) -> str:
    """
    Uses a language model to replace words with their synonyms.
    Args:
        all_prompts (List[str]): The input prompts as a list to perturb.
        model (str): The language model to use for generating synonyms as a vllm LLM object.
        instruction (str): The instruction prompt for the model.
    Returns:
        List[str]: The prompts with words replaced by their synonyms as a list of strings.
    """
    system_msg = (
        "You are a helpful assistant that replaces words with their synonyms "
        "while preserving the original meaning."
    )

    all_segments_to_perturb = []
    prompt_maps = []

    for prompt in all_prompts:
        parts = re.split(r"(\{.*?\})", prompt)
        structure = []  # Stores (is_placeholder, content)
        for part in parts:
            is_placeholder = part.startswith("{") and part.endswith("}")
            if not is_placeholder and part.strip():
                structure.append((False, len(all_segments_to_perturb)))
                all_segments_to_perturb.append(part)
            else:
                structure.append((True, part))
        prompt_maps.append(structure)

    flat_results, _, _ = batch_generation(
        model=model,
        system_messages=[system_msg] * len(all_segments_to_perturb),
        prompts=[instruction + text for text in all_segments_to_perturb],
        # response_generation_method=[ResponseGenerationMethod()] * len(all_segments_to_perturb),
        max_tokens=1024,
    )

    final_prompts = []
    for structure in prompt_maps:
        reconstructed = []
        for is_placeholder, content in structure:
            if is_placeholder:
                reconstructed.append(content)
            else:
                reconstructed.append(flat_results[content])
        final_prompts.append("".join(reconstructed))

    return final_prompts


def make_paraphrase(all_prompts: list[str], model: str, instruction: str) -> str:
    """
    Uses a language model to paraphrase the input text.
    Args:
        all_prompts (List[str]): The input prompts as a list to perturb.
        model (str): The language model to use for paraphrasing as a vllm LLM object.
        instruction (str): The instruction prompt for the model.
    Returns:
        List[str]: The paraphrased text as a list of strings.
    """
    system_msg = (
        "You are a helpful assistant that paraphrases text while preserving the original meaning."
    )

    all_segments_to_perturb = []
    prompt_maps = []

    for prompt in all_prompts:
        parts = re.split(r"(\{.*?\})", prompt)
        structure = []  # Stores (is_placeholder, content)
        for part in parts:
            is_placeholder = part.startswith("{") and part.endswith("}")
            if not is_placeholder and part.strip():
                structure.append((False, len(all_segments_to_perturb)))
                all_segments_to_perturb.append(part)
            else:
                structure.append((True, part))
        prompt_maps.append(structure)

    flat_results, _, _ = batch_generation(
        model=model,
        system_messages=[system_msg] * len(all_segments_to_perturb),
        prompts=[instruction + text for text in all_segments_to_perturb],
        # response_generation_method=[ResponseGenerationMethod()] * len(all_segments_to_perturb),
        max_tokens=1024,
    )

    final_prompts = []
    for structure in prompt_maps:
        reconstructed = []
        for is_placeholder, content in structure:
            if is_placeholder:
                reconstructed.append(content)
            else:
                reconstructed.append(flat_results[content])
        final_prompts.append("".join(reconstructed))

    return final_prompts


def apply_safe_perturbation(prompts: list, perturbation_func, **kwargs):
    """
    Splits list of prompts by curly brace placeholders (e.g., {PROMPT_OPTIONS}).
    Applies the perturbation_func ONLY to the prompts segments, protecting the keys.

    Args:
        prompts (List[str]): The input prompts containing placeholders.
        perturbation_func (function): The function to apply to non-placeholder text.
        **kwargs: Additional keyword arguments to pass to the perturbation
            function (e.g., probability).
    Returns:
        List[str]: The prompts with perturbations applied safely.
    """
    import re

    if not prompts:
        return prompts

    if perturbation_func in [make_synonyms, make_paraphrase]:
        print("Using batch perturbation function:", perturbation_func)
        final_prompts = prompts
        if perturbation_func == make_synonyms:
            final_prompts = make_synonyms(
                all_prompts=prompts,
                model=kwargs.get("model"),
                instruction=kwargs.get("instruction"),
            )
        elif perturbation_func == make_paraphrase:
            final_prompts = make_paraphrase(
                all_prompts=prompts,
                model=kwargs.get("model"),
                instruction=kwargs.get("instruction"),
            )
        return final_prompts
    else:
        perturbed_prompts = []
        for prompt in prompts:

            parts = re.split(r"(\{.*?\})", prompt)

            processed_parts = []
            for part in parts:
                # Check if this part is a placeholder
                if part.startswith("{") and part.endswith("}"):
                    # Append exactly as is
                    processed_parts.append(part)
                else:
                    # Apply the typo function
                    processed_parts.append(perturbation_func(part, **kwargs))

            perturbed_prompts.append("".join(processed_parts))

        return perturbed_prompts



from enum import Enum
from typing import Final

# Survey Item
SYSTEM_PROMPT_FIELD: Final[str] = "system_prompt"
QUESTIONNAIRE_INSTRUCTION_FIELD: Final[str] = "questionnaire_instruction"
QUESTIONNAIRE_ITEM_ID: Final[str] = "questionnaire_item_id"
QUESTIONNAIRE_ITEM: Final[str] = "questionnaire_item"
QUESTIONNAIRE_NAME: Final[str] = "questionnaire_name"

# Question
QUESTION_STEM: Final[str] = "question_stem"
QUESTION_CONTENT: Final[str] = "question_content"
QUESTION: Final[str] = "question"

# Answer
ANSWER_CODE: Final[str] = "answer_code"
ANSWER_TEXT: Final[str] = "answer_text"
ANSWER_OPTION: Final[str] = "answer_option"

# LLM Response
LLM_RESPONSE: Final[str] = "llm_response"
PARSED_RESPONSE: Final[str] = "parsed_response"


class QuestionnairePresentation(Enum):
    SINGLE_ITEM: str = "questionnaire_type_single_item"
    SEQUENTIAL: str = "questionnaire_type_sequential"
    BATTERY: str = "questionnaire_type_battery"


DEFAULT_SYSTEM_PROMPT: Final[str] = (
    "You will be given questions and possible answer options for each. "
    "Please reason about each question before answering."
)
DEFAULT_TASK_INSTRUCTION: Final[str] = ""
DEFAULT_JSON_STRUCTURE: Final[list[str]] = ["reasoning", "answer"]
DEFAULT_QUESTIONNAIRE_ID: Final[str] = "Questionnaire"



"""Tests for tree-based pydantic model generation used by structured outputs."""

import pytest
from pydantic import ValidationError

from qstn.inference.dynamic_pydantic import build_pydantic_model_from_json_object
from qstn.inference.response_generation import Constraints, JSONItem, JSONObject


def test_build_pydantic_model_enforces_enum_values():
    """Enum constraints should produce schema-backed runtime validation."""
    model = build_pydantic_model_from_json_object(
        JSONObject(
            children=[
                JSONItem(
                    json_field="color",
                    constraints=Constraints(enum=["red", "blue"]),
                )
            ]
        )
    )

    inst = model(color="red")

    assert inst.color.value == "red"
    with pytest.raises(ValidationError):
        model(color="green")


def test_build_pydantic_model_uses_declared_scalar_types():
    """Declared scalar types should be enforced by the generated model."""
    model = build_pydantic_model_from_json_object(
        JSONObject(children=[JSONItem(json_field="score", value_type="float")])
    )

    inst = model(score=0.5)

    assert isinstance(inst.score, float)
    with pytest.raises(ValidationError):
        model(score="not a float")


def test_build_pydantic_model_supports_nested_objects_and_aliases():
    """Nested JSONObjects should become nested pydantic models with aliases."""
    model = build_pydantic_model_from_json_object(
        JSONObject(
            children=[
                JSONObject(
                    json_field="Question 1",
                    children=[JSONItem(json_field="answer", value_type="int")],
                )
            ]
        )
    )

    inst = model(**{"Question 1": {"answer": 2}})

    assert inst.question_1.answer == 2
    with pytest.raises(ValidationError):
        model(**{"Question 1": {"answer": "wrong"}})


def test_build_pydantic_model_rejects_nested_objects_without_json_field():
    """Nested JSONObjects must expose a field name in the parent schema."""
    with pytest.raises(ValueError, match="Nested JSONObject entries must define `json_field`."):
        build_pydantic_model_from_json_object(
            JSONObject(children=[JSONObject(children=[JSONItem(json_field="answer")])])
        )



"""End-to-end smoke tests for core survey execution modes.

These tests validate that single-item, sequential, and battery survey flows
produce expected dataframe outputs with mocked API responses.
"""

import pandas as pd

import qstn
from qstn.utilities.survey_objects import QuestionnaireItem


def test_simple_single_item(
    mock_questionnaires, mock_personas, llm_prompt_factory, mock_openai_client
):
    # --- SETUP ---
    interviews = llm_prompt_factory(mock_personas, mock_questionnaires)

    # Add additional question only to the first interview
    interview = interviews[0]
    new_item = QuestionnaireItem(item_id=3, question_content="How do you feel about Green?")
    interview.insert_questions(new_item, 1)

    results = qstn.survey_manager.conduct_survey_single_item(
        model=mock_openai_client,
        llm_prompts=interviews,
        max_tokens=100,
        client_model_name="mock-model",
        api_concurrency=3,
    )

    # --- PARSING ---
    parsed_results = qstn.parser.raw_responses(results)

    # --- ASSERTIONS ---
    first_interview = parsed_results[interviews[0]]

    # 1. Check if we got a DataFrame back
    assert isinstance(first_interview, pd.DataFrame)
    assert not first_interview.empty

    # 2. Check logic for the specific interview we modified (First Persona)
    # We expect 3 answers because we started with 2 and inserted 1.
    assert len(first_interview) == 3

    # 3. Check if the INSERTED question is present and in the correct relative position
    # The IDs should be ordered 1 -> 3 -> 2 based on our insertion at index 1
    assert 3 in first_interview["questionnaire_item_id"].values
    assert "How do you feel about Green?" == first_interview["question"].iloc[1]

    # 4. Check if the mocked answer came through (just to test if it was actually called)
    assert first_interview.iloc[0]["llm_response"] == "I feel neutral about this."

    # 5. Check the second interview (which we did NOT touch)
    # It should still only have 2 questions and answers
    second_interview = parsed_results[interviews[1]]
    assert len(second_interview) == 2


def test_simple_sequential(
    mock_questionnaires, mock_personas, llm_prompt_factory, mock_openai_client
):
    # --- SETUP ---
    interviews = llm_prompt_factory(mock_personas, mock_questionnaires)

    # Add additional question only to the first interview
    interview = interviews[0]
    new_item = QuestionnaireItem(item_id=3, question_content="How do you feel about Green?")
    interview.insert_questions(new_item, 1)

    results = qstn.survey_manager.conduct_survey_sequential(
        model=mock_openai_client,
        llm_prompts=interviews,
        max_tokens=100,
        client_model_name="mock-model",
        api_concurrency=3,
    )

    # --- PARSING ---
    parsed_results = qstn.parser.raw_responses(results)

    # --- ASSERTIONS ---
    first_interview = parsed_results[interviews[0]]

    # 1. Check if we got a DataFrame back
    assert isinstance(first_interview, pd.DataFrame)
    assert not first_interview.empty

    # 2. Check logic for the specific interview we modified (First Persona)
    # We expect 3 answers because we started with 2 and inserted 1.
    assert len(first_interview) == 3

    # 3. Check if the INSERTED question is present and in the correct relative position
    # The IDs should be ordered 1 -> 3 -> 2 based on our insertion at index 1
    assert 3 in first_interview["questionnaire_item_id"].values
    assert "How do you feel about Green?" == first_interview["question"].iloc[1]

    # 4. Check if the mocked answer came through (just to test if it was actually called)
    assert first_interview.iloc[0]["llm_response"] == "I feel neutral about this."

    # 5. Check the second interview (which we did NOT touch)
    # It should still only have 2 questions and answers
    second_interview = parsed_results[interviews[1]]
    assert len(second_interview) == 2


def test_simple_battery(mock_questionnaires, mock_personas, llm_prompt_factory, mock_openai_client):
    # --- SETUP ---
    interviews = llm_prompt_factory(mock_personas, mock_questionnaires)

    # Add additional question only to the first interview
    interview = interviews[0]
    new_item = QuestionnaireItem(item_id=3, question_content="How do you feel about Green?")
    interview.insert_questions(new_item, 1)

    results = qstn.survey_manager.conduct_survey_battery(
        model=mock_openai_client,
        llm_prompts=interviews,
        max_tokens=100,
        client_model_name="mock-model",
        api_concurrency=3,
    )

    # --- PARSING ---
    parsed_results = qstn.parser.raw_responses(results)

    # --- ASSERTIONS ---
    first_interview = parsed_results[interviews[0]]

    # 1. Check if we got a DataFrame back
    assert isinstance(first_interview, pd.DataFrame)
    assert not first_interview.empty

    # 2. Battery mode aggregates into a single questionnaire row.
    assert len(first_interview) == 1
    assert first_interview["questionnaire_item_id"].iloc[0] == -1
    assert "How do you feel about Green?" in first_interview["question"].iloc[0]

    # 3. Check if the mocked answer came through.
    assert first_interview.iloc[0]["llm_response"] == "I feel neutral about this."



import streamlit as st


class StatefulWidgets:
    """
    A class to create Streamlit widgets with encapsulated session state management.
    """

    def __init__(self, prefix: str = "_"):
        """
        Initializes the StatefulWidgets class.

        Args:
            prefix (str): A prefix to use for the widget's internal session state key.
        """
        self.prefix = prefix

    def _store_value(self, key: str) -> None:
        """
        Callback function to store the widget's value from the internal state
        to the main session state.
        """
        st.session_state[key] = st.session_state[f"{self.prefix}{key}"]

    def _initialize_and_load(self, key: str, initial_value) -> None:
        """
        Initializes the session state for a given key if it doesn't exist,
        and then loads this value into the widget's internal state.
        """
        if key not in st.session_state:
            st.session_state[key] = initial_value
        st.session_state[f"{self.prefix}{key}"] = st.session_state[key]

    def create(self, widget_func, key: str, *args, initial_value=None, **kwargs):
        """
        Creates a Streamlit widget and binds its state.

        Args:
            widget_func: The Streamlit widget function (e.g., st.text_input).
            key (str): The key for the main session state.
            *args: Positional arguments to pass to the widget function.
            initial_value: The initial value to set in the session state if not present.
            **kwargs: Keyword arguments to pass to the widget function.

        Returns:
            The created Streamlit widget.
        """
        self._initialize_and_load(key, initial_value)
        return widget_func(
            *args,
            key=f"{self.prefix}{key}",
            on_change=self._store_value,
            args=(key,),
            **kwargs,
        )



import re


def parse_reasoning(full_text: str, patterns: list[str]):
    # If we have no reasoning, directly output everything
    extracted_reasoning = None
    final_answer = full_text

    for start_tag, end_tag in patterns:
        # --- Check if start tag and end tag in full_text --- #
        has_start = start_tag is not None and start_tag != "" and start_tag in full_text
        has_end = end_tag is not None and end_tag != "" and end_tag in full_text

        # --- CASE A: No Reasoning ---
        if not has_start and not has_end:
            continue

        # --- CASE B: Standard Case (Start ... End) ---
        if has_start and has_end:
            # Prepare regex safe patterns
            esc_start = re.escape(start_tag)
            esc_end = re.escape(end_tag)
            pattern_full = re.compile(f"{esc_start}(.*?){esc_end}", re.DOTALL)
            match = pattern_full.search(full_text)

            if match:
                extracted_reasoning = match.group(1)
                # Remove reasoning + tags from answer
                final_answer = pattern_full.sub("", full_text, count=1).strip()
        # --- CASE C: No Start tag in generated text. ---
        elif has_end and not has_start:
            parts = full_text.split(end_tag, 1)

            extracted_reasoning = parts[0].strip()
            # If output stops exactly at reasoning -> empty final answer
            if len(parts) > 1:
                final_answer = parts[1].strip()
            else:
                final_answer = ""

        # --- CASE D: Cut-off Case (Max Tokens Hit), No End Tag but a start tag ---
        # Some models do not have a start token (e.g. Qwen-3-VL).
        # In this case we cannot decide if it is reasoning or actual output
        elif has_start and not has_end:
            parts = full_text.split(start_tag, 1)
            extracted_reasoning = parts[1].strip()
            final_answer = ""  # Reasoning wasn't finished, so no answer exist

    return final_answer, extracted_reasoning



"""Unit tests for generic helper functions in `qstn.utilities.utils`.

These tests validate determinism, dataframe assembly behavior, and safe template
substitution semantics.
"""

import pandas as pd

from qstn.utilities import utils


class DummyKey:
    """Minimal key object exposing `questionnaire_name` for dataframe merge tests."""

    def __init__(self, name):
        self.questionnaire_name = name


def test_make_cache_key_stable():
    """`_make_cache_key` should be deterministic for identical inputs."""
    key1 = utils._make_cache_key([1,2], {"a":3})
    key2 = utils._make_cache_key([1,2], {"a":3})
    assert key1 == key2
    assert isinstance(key1, str)


def test_generate_seeds_reproducible():
    """`generate_seeds` should be reproducible and sized by batch."""
    seeds1 = utils.generate_seeds(seed=5, batch_size=3)
    seeds2 = utils.generate_seeds(seed=5, batch_size=3)
    assert seeds1 == seeds2
    assert len(seeds1) == 3
    # Different base seed should produce a different sample stream.
    seeds3 = utils.generate_seeds(seed=6, batch_size=3)
    assert seeds3 != seeds1


def test_create_one_dataframe_empty():
    """create_one_dataframe should return empty DataFrame when given empty dict."""
    assert utils.create_one_dataframe({}).empty


def test_create_one_dataframe_nonempty():
    """Merged dataframe should include source questionnaire names and all rows."""
    df1 = pd.DataFrame({"a":[1]})
    df2 = pd.DataFrame({"a":[2]})
    data = {
        DummyKey("Q1"): df1,
        DummyKey("Q2"): df2,
    }
    result = utils.create_one_dataframe(data)
    assert len(result) == 2
    assert "questionnaire_name" in result.columns
    assert set(result.questionnaire_name) == {"Q1","Q2"}


def test_safe_format_with_regex():
    """Regex formatter should replace known placeholders and keep unknown placeholders."""
    template = "Hello {{name}}, {{missing}} stays."
    out = utils.safe_format_with_regex(template, {"{{name}}":"Alice"})
    assert "Alice" in out
    assert "{{missing}}" in out


def test_safe_format_empty_template():
    """Formatting an empty template should return an empty string."""
    assert utils.safe_format_with_regex("", {}) == ""



import streamlit as st


def paginator(items, session_state_key):
    """
    Creates a reusable paginator component for a list of items.

    Args:
        items (list): The list of items to paginate through.
        session_state_key (str): A unique key for storing the current index in
                                 st.session_state.

    Returns:
        Any: The currently selected item from the list.
    """
    list_length = len(items)

    # Initialize the session state for the current index if it doesn't exist.
    if session_state_key not in st.session_state:
        st.session_state[session_state_key] = 0

    def next_item():
        """Increments the index, wrapping around to the start if it reaches the end."""
        if list_length > 0:
            st.session_state[session_state_key] = (
                st.session_state[session_state_key] + 1
            ) % list_length

    def prev_item():
        """Decrements the index, wrapping around to the end if it goes below zero."""
        if list_length > 0:
            st.session_state[session_state_key] = (
                st.session_state[session_state_key] - 1 + list_length
            ) % list_length

    # Create the navigation columns.
    col1, col2, col3 = st.columns([2, 3, 2])

    with col1:
        st.button(
            "⬅️ Previous",
            on_click=prev_item,
            use_container_width=True,
            disabled=(list_length <= 1),
        )

    # Display the current position and the popover for jumping to a specific item.
    with col2:
        if list_length > 0:
            current_num = st.session_state[session_state_key] + 1

            popover = st.popover(
                f"Item {current_num} of {list_length}",
                use_container_width=True,
            )

            with popover:
                st.markdown("Jump to a specific item:")
                target_index_input = st.number_input(
                    "Item Number",
                    min_value=1,
                    max_value=list_length,
                    value=current_num,
                    step=1,
                    label_visibility="collapsed",
                )
                if st.button("Go"):
                    st.session_state[session_state_key] = target_index_input - 1
                    st.rerun()
        else:
            st.text("No items to display")

    # "Next" button in the third column.
    with col3:
        st.button(
            "Next ➡️",
            on_click=next_item,
            use_container_width=True,
            disabled=(list_length <= 1),
        )

    # Return the currently selected item.
    if list_length > 0:
        return st.session_state[session_state_key]
    else:
        return None



import copy
import random
import warnings
from dataclasses import replace
from string import ascii_lowercase, ascii_uppercase
from typing import Any, Literal, Self, overload

import pandas as pd

from .inference.response_generation import (
    ResponseGenerationMethod,
    resolve_battery_response_generation_method,
)
from .utilities import constants, placeholder, prompt_templates
from .utilities.constants import QuestionnairePresentation
from .utilities.survey_objects import AnswerOptions, AnswerTexts, QuestionnaireItem
from .utilities.utils import safe_format_with_regex

# from transformers import AutoTokenizer


class LLMPrompt:
    """
    Main class for setting up and managing the prompt in the LLM experiment.

    This class handles loading questions
        from a predefined questionnaire, preparing prompts, managing answer options,
        and generating prompt structures for different interview types.
    """

    DEFAULT_QUESTIONNAIRE_ID: str = "Questionnaire"

    DEFAULT_SYSTEM_PROMPT: str = (
        "You will be given questions and possible answer options for each. "
        "Please reason about each question before answering."
    )
    DEFAULT_TASK_INSTRUCTION: str = ""

    DEFAULT_JSON_STRUCTURE: list[str] = ["reasoning", "answer"]

    DEFAULT_PROMPT_STRUCTURE: str = f"{placeholder.PROMPT_QUESTIONS}\n{placeholder.PROMPT_OPTIONS}"

    def __init__(
        self,
        questionnaire_source: str | pd.DataFrame = None,
        questionnaire_name: str = DEFAULT_QUESTIONNAIRE_ID,
        system_prompt: str | None = DEFAULT_SYSTEM_PROMPT,
        prompt: str = DEFAULT_PROMPT_STRUCTURE,
        verbose=False,
        seed: int = 42,
    ):
        """
        Initialize an LLMPrompt instance. Either a path to a csv file
            or a pandas dataframe can be provided to structure the questionnaire.
        Question structure can later be modified with explicit methods such as
        `insert_questions`, `replace_question`, and `remove_question`.

        Args:
            questionnaire_source (str/pd.Dataframe): Path to the CSV file containing the
                questionnaire structure and questions.
            questionnaire_name (str): Name/ID for the questionnaire.
            system_prompt (str | None): System prompt for all questions.
                Set to `None` to omit a system message.
            prompt (str): Prompt for all questions.
            verbose (bool): If True, enables verbose output. (Not implemented yet)
            seed (int): Random seed for reproducibility.
        """
        random.seed(seed)

        self._questions: list[QuestionnaireItem] = []

        if self._check_valid_questionnaire(questionnaire_source):
            self.load_questionnaire_format(questionnaire_source=questionnaire_source)

        self.verbose: bool = verbose

        self.questionnaire_name: str = questionnaire_name

        self.system_prompt: str | None = system_prompt
        self.prompt: str = prompt

    def _check_valid_questionnaire(self, questionnaire_source: str | pd.DataFrame = None) -> bool:
        # No Object
        if questionnaire_source is None:
            return False

        # Empty String
        if isinstance(questionnaire_source, str) and not questionnaire_source:
            return False

        # Empty Dataframe
        if isinstance(questionnaire_source, pd.DataFrame):
            if questionnaire_source.empty:
                warnings.warn(
                    "The provided Dataframe is empty! No questions are created.", stacklevel=2
                )
                return False
            # Optional check if the correct columns are provided?
            # Would probably be nice to have that warning here.

        return True

    def duplicate(self):
        """
        Create a deep copy of the current interview instance.

        Returns:
            LLMQuestionnaire: A deep copy of the current object.
        """
        return copy.deepcopy(self)

    def get_prompt_for_questionnaire_type(
        self,
        questionnaire_type: QuestionnairePresentation = QuestionnairePresentation.SINGLE_ITEM,
        item_id: str | int | None = None,
        item_position: int | None = 0,
        item_separator: str = "\n",
    ) -> tuple[str | None, str]:
        """
        Generate the full prompt for a given questionnaire presentation.

        Args:
            quesitonnaire_type (QuestionnairePresentation):
                The type of questionnaire prompt to generate.
            item_id (str):
                The id of the questionnaire_item that should be shown.
                If both item_id and item_position are provided, only item_id is considered.
            item_position (int): The question at that position will be shown.
                If both item_id and item_position are provided, only item_id is considered.
                Defaults to the first question.
            item_separator (str): For QuestionnairePresentation.BATTERY decides the str
                that seperates each question.
        Returns:
            Tuple(str | None, str): The first element corresponds to the system_prompt,
                the second element to the prompt.
        """
        options = ""
        automatic_output_instructions = ""

        question_map = {question.item_id: question for question in self._questions}
        reference_item_position = item_position
        if item_id:
            question_item = question_map[item_id]
            reference_item_position = next(
                i for i, question in enumerate(self._questions) if question.item_id == item_id
            )
        elif item_id and item_id not in question_map.keys():
            raise ValueError("item_id does not exist.")
        elif item_position >= len(self._questions):
            raise ValueError("item_order_id is bigger than the number of questions")
        else:
            question_item = self._questions[item_position]

        if (
            questionnaire_type == QuestionnairePresentation.SINGLE_ITEM
            or questionnaire_type == QuestionnairePresentation.SEQUENTIAL
        ):

            question = self.generate_question_prompt(question_item)

            if question_item.answer_options:
                options = question_item.answer_options.create_options_str()

                rgm = question_item.answer_options.response_generation_method
                if rgm is None:  # by default, no response generation method is required
                    automatic_output_instructions = ""
                else:
                    automatic_output_instructions: str = rgm.get_automatic_prompt()
            else:
                options = ""
                automatic_output_instructions = ""

            format_dict = {
                placeholder.PROMPT_QUESTIONS: question,
                placeholder.PROMPT_OPTIONS: options,
                placeholder.PROMPT_AUTOMATIC_OUTPUT_INSTRUCTIONS: automatic_output_instructions,
            }

        elif questionnaire_type == QuestionnairePresentation.BATTERY:
            all_questions: list[str] = []
            for question in self._questions:
                current_question_prompt = self.generate_question_prompt(question)

                if question.answer_options:
                    options = question.answer_options.create_options_str()
                else:
                    options = ""
                format_dict = {
                    placeholder.PROMPT_OPTIONS: options,
                }
                current_question_prompt = safe_format_with_regex(
                    current_question_prompt, format_dict
                )
                all_questions.append(current_question_prompt)

            all_questions_str = item_separator.join(all_questions)
            if question_item.answer_options:
                options = question_item.answer_options.create_options_str()
            else:
                options = ""

            rgm = resolve_battery_response_generation_method(
                questions=list(self._questions),
                item_position=reference_item_position,
            )
            if rgm is None:  # by default, no response generation method is required
                automatic_output_instructions = ""
            else:
                automatic_output_instructions = rgm.get_automatic_prompt()

            format_dict = {
                placeholder.PROMPT_QUESTIONS: all_questions_str,
                placeholder.PROMPT_OPTIONS: options,
                placeholder.PROMPT_AUTOMATIC_OUTPUT_INSTRUCTIONS: automatic_output_instructions,
            }

        if self.system_prompt is None:
            system_prompt = None
        else:
            system_prompt = safe_format_with_regex(self.system_prompt, format_dict)
        prompt = safe_format_with_regex(self.prompt, format_dict)

        return system_prompt, prompt

    # Currently disallowed to allow lightweight import. This should be reenabled in the future.
    # def calculate_input_token_estimate(
    #     self, model_id: str, questionnaire_type: QuestionnairePresentation =
    # QuestionnairePresentation.SINGLE_ITEM
    # ) -> int:
    #     """
    #     Estimate the number of input tokens for the prompt, given a model and questionnaire type.
    #     Remember that the model also has to have enough context length to fit its own response
    #     in case of CONTEXT and ONE_PROMPT type.

    #     Args:
    #         model_id (str): Huggingface model id.
    #         questionnaire_type (QuestionnairePresentation): Type of questionnaire prompt.

    #     Returns:
    #         int: Estimated number of input tokens.
    #     """
    #     tokenizer = AutoTokenizer.from_pretrained(model_id)
    #     system_prompt, prompt = self.get_prompt_for_questionnaire_type(
    #         questionnaire_type=questionnaire_type
    #     )
    #     system_tokens = tokenizer.encode(system_prompt)
    #     tokens = tokenizer.encode(prompt)
    #     total_tokens = len(system_tokens) + len(tokens)

    #     return (
    #         total_tokens
    #         if questionnaire_type != QuestionnairePresentation.SEQUENTIAL
    #         else len(total_tokens) * 3
    #     )

    def get_questions(self) -> tuple[QuestionnaireItem, ...]:
        """
        Get an immutable snapshot of loaded interview questions.

        Returns:
            Tuple[QuestionnaireItem, ...]: Loaded questions.
        """
        return tuple(self._questions)

    @property
    def questions(self) -> tuple[QuestionnaireItem, ...]:
        """Read-only view of questionnaire items."""
        return tuple(self._questions)

    def get_question(self, position: int) -> QuestionnaireItem:
        """Return a question by positional index."""
        return self._questions[position]

    def replace_question(self, position: int, questionnaire_item: QuestionnaireItem) -> None:
        """Replace the question at a given index."""
        self._questions[position] = questionnaire_item

    def remove_question(self, position: int) -> None:
        """Remove the question at a given index."""
        del self._questions[position]

    def get_question_item_id(self, position: int) -> Any:
        """Return the questionnaire item id at a given index."""
        return self._questions[position].item_id

    def load_questionnaire_format(self, questionnaire_source: str | pd.DataFrame) -> Self:
        """
        Load the questionnaire format from a CSV file or a pandas DataFrame.

        The CSV or pd.Dataframe must have the columns: questionnaire_item_id, question_content
        Optionally it can also have a question_stem.

        Args:
            questionnaire_source (str or pd.Dataframe): Path to a valid CSV file or pd.Dataframe.

        Returns:
            Self: The updated instance with loaded questions.
        """
        questionnaire_questions: list[QuestionnaireItem] = []

        # This is a duplicate check with actual Error here,
        # because if the method is called on its own it should not run the remaining code
        if not self._check_valid_questionnaire(questionnaire_source=questionnaire_source):
            raise ValueError("Please provide a non empty DataFrame or a valid String.")

        if isinstance(questionnaire_source, pd.DataFrame):
            df = questionnaire_source
        else:
            df = pd.read_csv(questionnaire_source)

        for _, row in df.iterrows():
            questionnaire_item_id = row[constants.QUESTIONNAIRE_ITEM_ID]
            # if constants.QUESTION in df.columns:
            #     question = row[constants.QUESTION]
            if constants.QUESTION_CONTENT in df.columns:
                questionnaire_question_content = row[constants.QUESTION_CONTENT]
            else:
                questionnaire_question_content = None

            if constants.QUESTION_STEM in df.columns:
                question_stem = row[constants.QUESTION_STEM]
            else:
                question_stem = None

            generated_questionnaire_question = QuestionnaireItem(
                item_id=questionnaire_item_id,
                question_content=questionnaire_question_content,
                question_stem=question_stem,
            )
            questionnaire_questions.append(generated_questionnaire_question)

        self._questions = questionnaire_questions
        return self

    # TODO Item order could be given by ids
    @overload
    def prepare_prompt(
        self,
        question_stem: str | None = None,
        answer_options: AnswerOptions | None = None,
        prefilled_responses: dict[int, str] | None = None,
        randomized_item_order: bool = False,
    ) -> Self: ...

    @overload
    def prepare_prompt(
        self,
        question_stem: list[str] | None = None,
        answer_options: dict[str, AnswerOptions] | None = None,
        prefilled_responses: dict[int, str] | None = None,
        randomized_item_order: bool = False,
    ) -> Self: ...

    def prepare_prompt(
        self,
        question_stem: str | list[str] | None = None,
        answer_options: AnswerOptions | dict[str, AnswerOptions] | None = None,
        prefilled_responses: dict[int, str] | None = None,
        randomized_item_order: bool = False,
    ) -> Self:
        """
        Prepare the interview by assigning question stems, answer options, and prefilled responses.

        Args:
            question_stem (str or List[str], optional): Single or list of question stems.
            answer_options (AnswerOptions or Dict[int, AnswerOptions], optional):
                Answer options for all or per question.
            prefilled_responses (Dict[int, str], optional):
                If you provide prefilled responses, they will be used
            to fill the answers instead of prompting the LLM for that question.
            randomized_item_order (bool): If True, randomize the order of questions.
        Returns:
            Self: The updated instance with prepared questions.
        """
        questionnaire_questions: list[QuestionnaireItem] = self._questions

        prompt_list = isinstance(question_stem, list)
        if prompt_list:
            assert len(question_stem) == len(
                questionnaire_questions
            ), "If a list of question stems is given, length of prompt "
            "       and survey questions have to be the same"

        options_dict = False

        if isinstance(answer_options, AnswerOptions):
            # self._same_options = True # unnecessary
            options_dict = False
        elif isinstance(answer_options, dict):
            # self._same_options = False # unnecessary
            options_dict = True

        updated_questions: list[QuestionnaireItem] = []

        if not prefilled_responses:
            prefilled_responses = {}
            # for survey_question in survey_questions:
            # prefilled_answers[survey_question.question_id] = None

        if not prompt_list and not options_dict:
            updated_questions = []
            for question in questionnaire_questions:
                new_questionnaire_question = replace(
                    question,
                    question_stem=(question_stem if question_stem else question.question_stem),
                    answer_options=answer_options,
                    prefilled_response=prefilled_responses.get(question.item_id),
                )
                updated_questions.append(new_questionnaire_question)

        elif not prompt_list and options_dict:
            for question in questionnaire_questions:
                new_questionnaire_question = replace(
                    question,
                    question_stem=(question_stem if question_stem else question.question_stem),
                    answer_options=answer_options.get(question.item_id),
                    prefilled_response=prefilled_responses.get(question.item_id),
                )
                updated_questions.append(new_questionnaire_question)

        elif prompt_list and not options_dict:
            for i, question in enumerate(questionnaire_questions):
                new_questionnaire_question = replace(
                    question,
                    question_stem=(question_stem[i] if question_stem else question.question_stem),
                    answer_options=answer_options,
                    prefilled_response=prefilled_responses.get(question.item_id),
                )
                updated_questions.append(new_questionnaire_question)
        elif prompt_list and options_dict:
            for i, question in enumerate(questionnaire_questions):
                new_questionnaire_question = replace(
                    question,
                    question_stem=(question_stem[i] if question_stem else question.question_stem),
                    answer_options=answer_options.get(question.item_id),
                    prefilled_response=prefilled_responses.get(question.item_id),
                )
                updated_questions.append(new_questionnaire_question)

        if randomized_item_order:
            random.shuffle(updated_questions)

        self._questions = updated_questions
        return self

    def generate_question_prompt(self, questionnaire_items: QuestionnaireItem) -> str:
        """
        Generate the prompt string for a single interview question.

        Args:
            questionnaire_items (InterviewItem): The question to prompt.

        Returns:
            str: The formatted prompt for the question.
        """

        if questionnaire_items.question_stem:
            if placeholder.QUESTION_CONTENT in questionnaire_items.question_stem:
                format_dict = {placeholder.QUESTION_CONTENT: questionnaire_items.question_content}
                question_prompt = safe_format_with_regex(
                    questionnaire_items.question_stem, format_dict
                )
            else:
                question_prompt = f"""{questionnaire_items.question_stem} {questionnaire_items.question_content}"""  # noqa: E501
        else:
            question_prompt = f"""{questionnaire_items.question_content}"""

        if questionnaire_items.answer_options:
            _options_str = questionnaire_items.answer_options.create_options_str()
            if _options_str is not None:
                safe_formatter = {placeholder.PROMPT_OPTIONS: _options_str}
                question_prompt = safe_format_with_regex(question_prompt, safe_formatter)

        return question_prompt

    def __len__(self) -> int:
        """
        Returns the number of questions in our LLMPrompt.

        Returns:
            int: The number of questions.
        """
        return len(self._questions)

    def __str__(self) -> str:
        """
        Creates a human readable display of the system prompt and prompt in default Battery format.
        """
        name_str: str = f"=== {self.questionnaire_name} ==="
        sys_prompt, prompt = self.get_prompt_for_questionnaire_type(
            questionnaire_type=QuestionnairePresentation.BATTERY
        )
        sys_str: str = f"=== SYSTEM_PROMPT ===\n{sys_prompt}"
        prompt_str: str = f"=== USER_PROMPT_WITH_ALL_QUESTIONS ===\n{prompt}"

        full_str: str = f"{name_str}\n{sys_str}\n{prompt_str}"
        return full_str

    def insert_questions(
        self,
        items: QuestionnaireItem | list[QuestionnaireItem],
        position: int = None,
    ) -> None:
        """Inserts one or more questions into the questionnaire.

        Args:
            items (Union[QuestionnaireItem, List[QuestionnaireItem]]): A single
                QuestionnaireItem or a list of items to insert.
            position (int): The index where the questions should be inserted.
                Default [None] adds them at the end.
        """
        if position is None:
            position = len(self._questions)

        if not isinstance(items, (list, tuple)):
            items = [items]

        self._questions[position:position] = items


_IDX_TYPES = Literal["char_lower", "char_upper", "integer", "no_index"]


def generate_likert_options(
    n: int,
    answer_texts: list[str] | None,
    only_from_to_scale: bool = False,
    random_order: bool = False,
    reversed_order: bool = False,
    even_order: bool = False,
    add_middle_category: bool = False,
    str_middle_cat: str = "Neutral",
    add_refusal: bool = False,
    refusal_code: str = "-99",
    start_idx: int = 1,
    list_prompt_template: str = prompt_templates.LIST_OPTIONS_DEFAULT,
    scale_prompt_template: str = prompt_templates.SCALE_OPTIONS_DEFAULT,
    index_answer_separator: str = ": ",
    options_separator: str = ", ",
    idx_type: _IDX_TYPES = "integer",
    response_generation_method: ResponseGenerationMethod | None = None,
) -> AnswerOptions:
    """Generates a set of options and a prompt for a Likert-style scale.

    This function creates a numeric or alphabetic scale of a specified size (n),
    optionally attaching textual labels to the scale. It provides
    extensive control over ordering, formatting, and the final prompt string.

    Args:
        n (int): The number of options to generate (e.g., 5 for a 5-point scale).
        answer_texts (Optional[List[str]]): A list of text labels for each option.
            Its length must equal `n` if provided.
        only_from_to_scale (bool, optional): If True, the prompt will only show the
            min and max of the scale (e.g., "1 to 5"). Defaults to False.
        random_order (bool, optional): If True, the options are randomized. Defaults to False.
        reversed_order (bool, optional): If True, the options are in reversed input order.
            Defaults to False.
        even_order (bool, optional): If True, options the center option will be removed.
            E.g., for n=5: 1, 2, 4, 5
        add_middle_category (bool, optional): If True, a middle category will be added.
            The name can be specified,
            by default it is "Neutral". E.g., for n=4: 1, 2, 3: Neutral, 4, 5
        str_middle_cat (str, optional): The label for the middle category
            if `add_middle_category` is True.
            Defaults to "Neutral".
        add_refusal (bool, optional): If True, an additional option for
            "Don't know / Refuse to answer" will be added.
            Defaults to False.
        refusal_code (str, optional): The code assigned to the refusal option
            if `add_refusal` is True.
            Defaults to "-99".
        start_idx (int, optional): The starting index for the scale (usually 0 or 1).
            Defaults to 1.
        list_prompt_template (str, optional): The template for prompts that list all options.
        scale_prompt_template (str, optional): The template for prompts that only show the range.
        index_answer_separator (str, optional): The string used to separate an index from its
            text label (e.g., "1: Strongly Agree"). Defaults to ": ".
        options_separator (str, optional): The string used to separate options when listed
            in the prompt. Defaults to ", ".
        idx_type (_IDX_TYPES, optional): The type of index to use: "integer", "upper" (A, B, C),
            or "lower" (a, b, c). Defaults to "integer".
        response_generation_method (Optional[ResponseGenerationMethod], optional): An object
            controlling how the final response object is generated. Defaults to None.

    Raises:
        ValueError: If `answer_texts` is provided and its length does not match `n`.

    Returns:
        AnswerOptions: An object containing the generated list of option strings and the
        final formatted prompt ready for display.

    Example:
        .. code-block:: python

            # Generate a classic 5-point "Strongly Disagree" to "Strongly Agree" scale
            labels = [
                "Strongly Disagree", "Disagree", "Neutral", "Agree", "Strongly Agree"
            ]
            options = SurveyOptionGenerator.generate_likert_options(n=5, answer_texts=labels)
    """

    if only_from_to_scale:
        # if len(answer_texts) != 2:
        #     raise ValueError(
        #         "From-To scales require exactly 2 descriptions, but "
        #         f"answer_texts was set to '{answer_texts}'."
        #     )
        if idx_type != "integer":
            raise ValueError(
                "From-To scales require an integer scale index, but "
                f"idx_type was set to '{idx_type}'."
            )
    else:
        if answer_texts:
            if len(answer_texts) != n:
                raise ValueError(
                    "answer_texts and n need to be the same length, but "
                    f"answer_texts has length {len(answer_texts)} "
                    f"and n was given as {n}."
                )
    if even_order:
        if n % 2 == 0:
            raise ValueError("If you want to turn a scale even, it should be odd before.")
        middle_index = n // 2
        answer_texts = answer_texts[:middle_index] + answer_texts[middle_index + 1 :]
        n = n - 1
    if add_middle_category:
        if n % 2 != 0:
            raise ValueError("If you want to add a middle category, it should be even before.")
        middle_index = n // 2
        answer_texts = answer_texts[:middle_index] + [str_middle_cat] + answer_texts[middle_index:]
        n = n + 1

    if random_order:
        if len(answer_texts) < 2:
            raise ValueError("There must be at least two answer options to reorder randomly.")
        random.shuffle(answer_texts)  # no assignment needed because shuffles already inplace
    if reversed_order:
        if len(answer_texts) < 2:
            raise ValueError("There must be at least two answer options to reorder in reverse.")
        answer_texts = answer_texts[::-1]

    if add_refusal:
        answer_texts.append("Don't know / Refuse to answer")
        n += 1

    answer_option_indices = []
    if idx_type == "no_index":
        # no index, just the answer options directly
        answer_option_indices = None
    elif idx_type == "integer":
        if add_refusal:  # if refusal is added, assign it a common code -99
            for i in range(n - 1):
                answer_code = i + start_idx
                answer_option_indices.append(str(answer_code))
            answer_option_indices.append(refusal_code)  # common code for refusal
        else:
            for i in range(n):
                answer_code = i + start_idx
                answer_option_indices.append(str(answer_code))
    else:
        # TODO @Jens add these to constants.py
        if idx_type == "char_lower":
            for i in range(n):
                answer_option_indices.append(ascii_lowercase[(i + start_idx) % 26])
        elif idx_type == "char_upper":
            for i in range(n):
                answer_option_indices.append(ascii_uppercase[(i + start_idx) % 26])

    answer_texts_object = AnswerTexts(
        answer_texts=answer_texts,
        indices=answer_option_indices,
        index_answer_seperator=index_answer_separator,
        option_seperators=options_separator,
        only_scale=only_from_to_scale,
    )

    questionnaire_options = AnswerOptions(
        answer_texts=answer_texts_object,
        from_to_scale=only_from_to_scale,
        list_prompt_template=list_prompt_template,
        scale_prompt_template=scale_prompt_template,
        response_generation_method=response_generation_method,
    )

    return questionnaire_options
